import ecl
import ecl.util.util
import ecl.util.geometry

from cwrap import Prototype

from .well_type_enum import WellTypeEnum
from .well_connection_direction_enum import WellConnectionDirectionEnum
from .well_connection import WellConnection
from .well_segment import WellSegment
from .well_state import WellState
from .well_time_line import WellTimeLine
from .well_info  import WellInfo



#  Copyright (C) 2018  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

"""
  ecl_grav/EclGrav: Class used to simplify evaluation of ECLIPSE
     modelling time-lapse gravitational surveys.

  ecl_subsidence/EclSubsidence: Small class used to evaluate simulated
     subsidence from ECLIPSE simulations; analogous to the EcLGrav
     functionality.
"""

from .ecl_subsidence import EclSubsidence
from .ecl_grav_calc import phase_deltag, deltag
from .ecl_grav import EclGrav



from cwrap import BaseCEnum

class WellConnectionDirectionEnum(BaseCEnum):
    TYPE_NAME = "well_connection_dir_enum"
    well_conn_dirX  = None
    well_conn_dirY  = None
    well_conn_dirZ  = None
    well_conn_fracX = None
    well_conn_fracY = None

WellConnectionDirectionEnum.addEnum("well_conn_dirX", 1)
WellConnectionDirectionEnum.addEnum("well_conn_dirY", 2)
WellConnectionDirectionEnum.addEnum("well_conn_dirZ", 3)
WellConnectionDirectionEnum.addEnum("well_conn_fracX", 4)
WellConnectionDirectionEnum.addEnum("well_conn_fracY", 5)



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCClass

from ecl.util.util import monkey_the_camel
from ecl.util.util import CTime
from ecl import EclPrototype



class EclSumTStep(BaseCClass):
    TYPE_NAME = "ecl_sum_tstep"
    _alloc        = EclPrototype("void* ecl_sum_tstep_alloc_new(int, int, float, void*)", bind=False)
    _free         = EclPrototype("void ecl_sum_tstep_free(ecl_sum_tstep)")
    _get_sim_days = EclPrototype("double ecl_sum_tstep_get_sim_days(ecl_sum_tstep)")
    _get_sim_time = EclPrototype("time_t ecl_sum_tstep_get_sim_time(ecl_sum_tstep)")
    _get_report   = EclPrototype("int ecl_sum_tstep_get_report(ecl_sum_tstep)")
    _get_ministep = EclPrototype("int ecl_sum_tstep_get_ministep(ecl_sum_tstep)")
    _set_from_key = EclPrototype("void ecl_sum_tstep_set_from_key(ecl_sum_tstep, char*, float)")
    _get_from_key = EclPrototype("double ecl_sum_tstep_get_from_key(ecl_sum_tstep, char*)")
    _has_key = EclPrototype("bool ecl_sum_tstep_has_key(ecl_sum_tstep, char*)")



    def __init__(self, report_step, mini_step, sim_days, smspec):
        sim_seconds = sim_days * 24 * 60 * 60
        c_pointer = self._alloc(report_step, mini_step, sim_seconds, smspec)
        super(EclSumTStep, self).__init__(c_pointer)


    def get_sim_days(self):
        """ @rtype: double """
        return self._get_sim_days()

    def get_report(self):
        """ @rtype: int """
        return self._get_report()

    def get_mini_step(self):
        """ @rtype: int """
        return self._get_ministep()

    def get_sim_time(self):
        """ @rtype: CTime """
        return self._get_sim_time()

    def __getitem__(self, key):
        """ @rtype: double """
        if not key in self:
            raise KeyError("Key '%s' is not available." % key)

        return self._get_from_key(key)

    def __setitem__(self, key, value):
        if not key in self:
            raise KeyError("Key '%s' is not available." % key)

        self._set_from_key(key, value)

    def __contains__(self, key):
        return self._has_key(key)

    def free(self):
        self._free(self)

    def __repr__(self):
        d = self._get_sim_days()
        t = self._get_sim_time()
        r = self._get_report()
        m = self._get_ministep()
        cnt = 'sim_days={}, sim_time={}, report={}, ministep={}'
        return self._create_repr(cnt.format(d, t, r, m))


monkey_the_camel(EclSumTStep, 'getSimDays', EclSumTStep.get_sim_days)
monkey_the_camel(EclSumTStep, 'getReport', EclSumTStep.get_report)
monkey_the_camel(EclSumTStep, 'getMiniStep', EclSumTStep.get_mini_step)
monkey_the_camel(EclSumTStep, 'getSimTime', EclSumTStep.get_sim_time)



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'ecl_3dkw.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from ecl.util.util import monkey_the_camel
from .ecl_kw import EclKW

class Ecl3DKW(EclKW):
    """
    Class for working with Eclipse keywords defined over a grid

    The Ecl3DKW class is derived from the EclKW class, and most of the
    methods are implemented in the EclKW base class. The purpose of
    the Ecl3DKW class is to simplify working with 3D properties like
    PORO or SATNUM.

    The Ecl3DKW class has an attached EclGrid which is used to support
    [i,j,k] indexing, and a defined default value which is used when
    reading an inactive value. The Ecl3DKW keyword instances are
    returned from the EclInitFile and EclRestartFile classes, in
    addition you can excplicitly \"cast\" a EclKW keyword to Ecl3DKW
    with the Ecl3DKW.castFromKW() classmethod.

    Usage example:

       from ecl.ecl import EclInitFile,EclGrid

       grid = EclGrid("ECLIPSE.EGRID")
       file = EclInitFile(grid , "ECLIPSE.INIT")

       permx_kw = file["PORO"][0]
       porv_kw = file["PORV"][0]

       permx_kw.setDefault( -1 )
       for k in range(grid.getNZ()):
           for j in range(grid.getNY()):
               for i in range(grid.getNX()):
                   print('"(%d,%d,%d)  Permx:%g  Porv:%g"' % (i,j,k,permx_kw[i,j,k] , porv_kw[i,j,k]))

    In the example we open an ECLIPSE INIT file and extract the PERMX
    and PORV properties, and then iterate over all the cells in the
    grid.

    In the INIT file the PORV keyword is stored with all cells,
    whereas the PERMX keyword typically only has the active cells
    stored, this active/inactive gymnastics is handled
    transparently. With the call:

        permx_kw.setDefault( -1 )

    we say that we want the value -1 for all inactive cells in the
    PERMX property.

    """
    def __init__(self, kw , grid , value_type , default_value = 0 , global_active = False):
        if global_active:
            size = grid.getGlobalSize()
        else:
            size = grid.getNumActive( )
        super(Ecl3DKW , self).__init__( kw , size , value_type)
        self.grid = grid
        self.global_active = global_active
        self.setDefault( default_value )


    @classmethod
    def create(cls , kw , grid , value_type , default_value = 0 , global_active = False):
        new_kw = Ecl3DKW(kw , grid , value_type , default_value , global_active)
        return new_kw

    @classmethod
    def read_grdecl( cls , grid , fileH , kw , strict = True , ecl_type = None):
        """
        Will load an Ecl3DKW instance from a grdecl formatted filehandle.

        See the base class EclKW.read_grdecl() for more documentation.
        """
        kw = super(Ecl3DKW , cls).read_grdecl( fileH , kw , strict , ecl_type)
        Ecl3DKW.castFromKW(kw , grid)
        return kw



    def __getitem__(self , index):
        """Will return item [g] or [i,j,k].

        The __getitem__() methods supports both scalar indexing like
        [g] and tuples [i,j,k]. If the input argument is given as a
        [i,j,k] tuple it is converted to an active index before the
        final lookup.

        If the [i,j,k] input corresponds to an inactive cell in a
        keyword with only nactive elements the default value will be
        returned. By default the default value will be 0, but another
        value can be assigned with the setDefault() method.
        """
        if isinstance(index , tuple):
            global_index = self.grid.get_global_index( ijk = index )
            if self.global_active:
                index = global_index
            else:
                if not self.grid.active( global_index = global_index):
                    return self.getDefault()
                else:
                    index = self.grid.get_active_index( ijk = index )


        return super(Ecl3DKW , self).__getitem__( index )




    def __setitem__(self , index , value):
        """Set the value of at index [g] or [i,j,k].

        The __setitem__() methods supports both scalar indexing like
        [g] and tuples [i,j,k]. If the input argument is given as a
        [i,j,k] tuple it is converted to an active index before the
        final assignment.

        If you try to assign an inactive cell in a keyword with only
        nactive elements a ValueError() exception will be raised.
        """
        if isinstance(index , tuple):
            global_index = self.grid.get_global_index( ijk = index )
            if self.global_active:
                index = global_index
            else:
                if not self.grid.active( global_index = global_index):
                    raise ValueError("Tried to assign value to inactive cell: (%d,%d,%d)" % index)
                else:
                    index = self.grid.get_active_index( ijk = index )


        return super(Ecl3DKW , self).__setitem__( index , value )


    @classmethod
    def cast_from_kw(cls, kw, grid, default_value=0):
        """Will convert a normal EclKW to a Ecl3DKW.

        The method will convert a normal EclKW instance to Ecl3DKw
        instance with an attached grid and a default value.

        The method will check that size of the keyword is compatible
        with the grid dimensions, i.e. the keyword must have either
        nactive or nx*ny*nz elements. If the size of the keyword is
        not compatible with the grid dimensions a ValueError exception
        is raised.

        Example:

          1. Load the poro keyword from a grdecl formatted file.
          2. Convert the keyword to a 3D keyword.


        from ecl.ecl import EclGrid,EclKW,Ecl3DKW

        grid = EclGrid("ECLIPSE.EGRID")
        poro = EclKW.read_grdecl(open("poro.grdecl") , "PORO")
        Ecl3DKW.castFromKW( poro , grid )

        print('Porosity in cell (10,11,12):%g' % poro[10,11,12])
        """
        if len(kw) == grid.getGlobalSize():
            kw.global_active = True
        elif len(kw) == grid.getNumActive():
            kw.global_active = False
        else:
            raise ValueError("Size mismatch - must have size matching global/active size of grid")


        kw.__class__ = cls
        kw.default_value = default_value
        kw.grid = grid
        if len(kw) == grid.getGlobalSize():
            kw.global_active = True
        else:
            kw.global_active = False

        kw.setDefault( default_value )
        return kw


    def compressed_copy(self):
        """Will return a EclKW copy with nactive elements.

        The returned copy will be of type EclKW; i.e. no default
        interpolation and only linear access in the [] operator. The
        main purpose of this is to facilitate iteration over the
        active index, and for writing binary files.
        """
        return self.grid.compressedKWCopy( self )


    def global_copy(self):
        """Will return a EclKW copy with nx*ny*nz elements.

        The returned copy will be of type EclKW; i.e. no default
        interpolation and only linear access in the [] operator. The
        main purpose of this is to facilitate iteration over the
        global index, and for writing binary files.
        """
        return self.grid.globalKWCopy( self , self.getDefault() )



    def dims(self):
        return (self.grid.getNX() , self.grid.getNY() , self.grid.getNZ())


    def set_default(self, default_value):
        self.default_value = default_value


    def get_default(self):
        return self.default_value


monkey_the_camel(Ecl3DKW, 'castFromKW', Ecl3DKW.cast_from_kw, classmethod)
monkey_the_camel(Ecl3DKW, 'compressedCopy', Ecl3DKW.compressed_copy)
monkey_the_camel(Ecl3DKW, 'globalCopy', Ecl3DKW.global_copy)
monkey_the_camel(Ecl3DKW, 'setDefault', Ecl3DKW.set_default)
monkey_the_camel(Ecl3DKW, 'getDefault', Ecl3DKW.get_default)



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl import EclPrototype

__arglist  = 'double, double, double, '
__arglist += 'ecl_grid, ecl_file, '
__arglist += 'ecl_kw, ecl_kw, ecl_kw, ecl_kw, ecl_kw, ecl_kw'
_phase_deltag = EclPrototype("double ecl_grav_phase_deltag(%s)" % __arglist)

def phase_deltag(xyz, grid, init, sat1, rho1, porv1, sat2, rho2, porv2):
    return _phase_deltag(xyz[0], xyz[1], xyz[2],
                         grid.c_ptr, init.c_ptr,
                         sat1.c_ptr, rho1.c_ptr, porv1.c_ptr,
                         sat2.c_ptr, rho2.c_ptr, porv2.c_ptr)


def deltag(xyz, grid, init_file, restart_file1, restart_file2):
    """
    1. All restart files should have water, i.e. the SWAT keyword.
    2. All phases present in the restart file should also be present as densities,
       in addition the model must contain one additional phase - which should have a density.
    3. The restart files can never contain oil saturation.
    """

    swat1 = restart_file1.iget_named_kw("SWAT", 0)
    swat2 = restart_file2.iget_named_kw("SWAT", 0)

    phase_list = [(swat1, swat2)]

    if restart_file1.has_kw("SGAS"):
        # This is a three phase Water / Gas / Oil system
        sgas1 = restart_file1.iget_named_kw("SGAS", 0)
        sgas2 = restart_file2.iget_named_kw("SGAS", 0)

        soil1 = 1 - (sgas1 + swat1)
        soil2 = 1 - (sgas2 + swat2)
        soil1.name = "SOIL"
        soil2.name = "SOIL"
        phase_list += [(sgas1, sgas2),
                       (soil1, soil2)]
    else:
        # This is a two phase Water / xxx System. Will look for
        # OIL_DEN and GAS_DEN keywords to determine whether it is a
        # Water / Oil or Water / Gas system.

        if restart_file1.has_kw("OIL_DEN"):
            # Oil / Water system
            soil1 = 1 - swat1
            soil2 = 1 - swat2
            soil1.name = "SOIL"
            soil2.name = "SOIL"
            phase_list += [(soil1, soil2)]
        else:
            # Gas / Water system
            sgas1 = 1 - swat1
            sgas2 = 1 - swat2
            sgas1.name = "SGAS"
            sgas2.name = "SGAS"
            phase_list += [(sgas1, sgas2)]

    porv1 = restart_file1.iget_named_kw("RPORV", 0)
    porv2 = restart_file2.iget_named_kw("RPORV", 0)

    deltag = 0
    for (sat1, sat2) in phase_list:
        rho_name = "%s_DEN" % sat1.name[1:]
        rho1 = restart_file1.iget_named_kw(rho_name, 0)
        rho2 = restart_file2.iget_named_kw(rho_name, 0)
        deltag += phase_deltag(xyz, grid, init_file, sat1, rho1, porv1, sat2, rho2, porv2)
    return deltag



import os

class CWDContext(object):
    def __init__(self , path):
        self.cwd = os.getcwd()
        if os.path.isdir( path ):
            os.chdir( path )
        else:
            raise IOError("Path:%s does not exist" % path)

    def __exit__(self , exc_type , exc_val , exc_tb):
        os.chdir( self.cwd )
        return False

    def __enter__(self):
        return self



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file '__init__.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Package with utility classes, used by other ERT classes.

The libutil library implements many utility functions and classes of
things like hash table and vector; these classes are extensively used
by the other ert libraries. The present wrapping here is to facilitate
use and interaction with various ert classes, in a pure python context
you are probably better served by using a plain python solution;
either based on built in python objects or well established third
party packages.

The modules included in the util package are:

  tvector.py: This module implements the classes IntVector,
     DoubleVector and BoolVector. This is a quite normal
     implementation of a typed growable vector; but with a special
     twist regarding default values.

  util_func.py: This module wraps a couple of stateless (i.e. there is
     no class involved) functions from the util.c file.

"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import ecl
from cwrap import Prototype



from .version import Version, EclVersion

from ecl.util.enums import RngAlgTypeEnum, RngInitModeEnum

from .ctime import CTime

from .permutation_vector import PermutationVector
from .vector_template import VectorTemplate
from .double_vector import DoubleVector
from .int_vector import IntVector
from .bool_vector import BoolVector
from .time_vector import TimeVector
from .stringlist import StringList
from .rng import RandomNumberGenerator
from .lookup_table import LookupTable
from .hash import Hash, StringHash, DoubleHash, IntegerHash
from .thread_pool import ThreadPool
from .install_abort_signals import installAbortSignals, updateAbortSignals
from .cwd_context import CWDContext



###
###  monkey_the_camel is a function temporarily added to libecl while we are in
###  the process of changing camelCase function names to snake_case function
###  names.
###
###  See https://github.com/Statoil/libecl/issues/142 for a discussion and for
###  usage.
###

import os
import warnings

__cc = os.environ.get('ECLWARNING', None)  # __cc in (None, 'user', 'dev', 'hard')

def __silencio(msg):
    pass

def __user_warning(msg):
    print('User warning: ' + msg)

def __dev_warning(msg):
    warnings.warn(msg, DeprecationWarning)

def __hard_warning(msg):
    raise UserWarning('CamelCase exception: ' + msg)


__ecl_camel_case_warning = __silencio
if __cc == 'user':
    __ecl_camel_case_warning = __user_warning
elif __cc == 'dev':
    __ecl_camel_case_warning = __dev_warning
elif __cc == 'hard':
    __ecl_camel_case_warning = __hard_warning


def monkey_the_camel(class_, camel, method_, method_type=None):
    """Creates a method "class_.camel" in class_ which prints a warning and forwards
    to method_.  method_type should be one of (None, classmethod, staticmethod),
    and generates new methods accordingly.
    """
    def shift(*args):
        return args if (method_type != classmethod) else args[1:]
    def warned_method(*args, **kwargs):
        __ecl_camel_case_warning('Warning, %s is deprecated, use %s' % (camel, str(method_)))
        return method_(*shift(*args), **kwargs)
    if method_type == staticmethod:
        warned_method = staticmethod(warned_method)
    elif method_type == classmethod:
        warned_method = classmethod(warned_method)
    setattr(class_, camel, warned_method)



#  Copyright (C) 2016  Statoil ASA, Norway.
#
#  The file 'surface' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
from __future__ import division

"""
Create a polygon
"""
import os.path
import ctypes

from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.util.geometry import GeoPointset


class Surface(BaseCClass):
    TYPE_NAME = "surface"

    _alloc        = EclPrototype("void*  geo_surface_fload_alloc_irap( char* , bool )" , bind = False)
    _free         = EclPrototype("void   geo_surface_free( surface )")
    _new          = EclPrototype("void*  geo_surface_alloc_new( int, int, double, double, double, double, double )", bind = False)
    _get_nx       = EclPrototype("int    geo_surface_get_nx( surface )")
    _get_ny       = EclPrototype("int    geo_surface_get_ny( surface )")
    _iget_zvalue  = EclPrototype("double geo_surface_iget_zvalue( surface , int)")
    _iset_zvalue  = EclPrototype("void   geo_surface_iset_zvalue( surface , int , double)")
    _write        = EclPrototype("void   geo_surface_fprintf_irap( surface , char* )")
    _equal        = EclPrototype("bool   geo_surface_equal( surface , surface )")
    _header_equal = EclPrototype("bool   geo_surface_equal_header( surface , surface )")
    _copy         = EclPrototype("surface_obj geo_surface_alloc_copy( surface , bool )")
    _assign       = EclPrototype("void   geo_surface_assign_value( surface , double )")
    _scale        = EclPrototype("void   geo_surface_scale( surface , double )")
    _shift        = EclPrototype("void   geo_surface_shift( surface , double )")
    _iadd         = EclPrototype("void   geo_surface_iadd( surface , surface )")
    _imul         = EclPrototype("void   geo_surface_imul( surface , surface )")
    _isub         = EclPrototype("void   geo_surface_isub( surface , surface )")
    _isqrt        = EclPrototype("void   geo_surface_isqrt( surface )")
    _iget_xy      = EclPrototype("void   geo_surface_iget_xy(surface, int, double*, double*)")
    _get_pointset = EclPrototype("geo_pointset_ref geo_surface_get_pointset(surface)")


    def __init__(self, filename=None, nx=None, ny=None, xinc=None, yinc=None,
                                      xstart=None, ystart=None, angle=None):
        """
        This will load a irap surface from file. The surface should
        consist of a header and a set z values.
        """
        if filename is not None:
            filename = str(filename)
            if os.path.isfile( filename ):
                c_ptr = self._alloc(filename , True)
                super(Surface , self).__init__(c_ptr)
            else:
                raise IOError('No such file "%s".' % filename)
        else:
            s_args = [nx, ny, xinc, yinc, xstart, ystart, angle]
            if None in s_args:
                raise ValueError('Missing argument for creating surface, all values must be set, was: %s' % str(s_args))
            c_ptr = self._new(*s_args)
            super(Surface , self).__init__(c_ptr)

    def __eq__(self , other):
        """
        Compares two Surface instances, both header and data must be equal
        to compare as equal.
        """
        if isinstance( other , Surface):
            return self._equal(other)
        else:
            return False


    def headerEqual(self , other):
        return self._header_equal( other)


    def __iadd__(self , other):
        if isinstance(other , Surface):
            if self.headerEqual(other):
                self._iadd(other)
            else:
                raise ValueError("Tried to add incompatible surfaces")
        else:
            self._shift(other)
        return self


    def __isub__(self , other):
        if isinstance(other , Surface):
            if self.headerEqual(other):
                self._isub(other)
            else:
                raise ValueError("Tried to subtract incompatible surfaces")
        else:
            self._shift( -other)
        return self


    def __imul__(self , other):
        if isinstance(other , Surface):
            if self.headerEqual(other):
                self._imul( other)
            else:
                raise ValueError("Tried to add multiply ncompatible surfaces")
        else:
            self._scale( other)
        return self

    def __itruediv__(self , other):
        self._scale(1.0 / other)
        return self

    def __idiv__(self, other):
        return self.__itruediv__(other)


    def __add__(self , other):
        copy = self.copy()
        copy += other
        return copy


    def __mul__(self , other):
        copy = self.copy()
        copy *= other
        return copy


    def __sub__(self , other):
        copy = self.copy()
        copy -= other
        return copy


    def __truediv__(self , other):
        copy = self.copy()
        copy /= other
        return copy


    def __div__(self, other):
        return self.__truediv__(other)


    def __len__(self):
        """
        The number of values in the surface.
        """
        return self.getNX() * self.getNY()


    def inplaceSqrt(self):
        """
        Will do an inplace sqrt operation.
        """
        self._isqrt( )
        return self


    def sqrt(self):
        """
        Will return a new surface where all elements have been sqrt{ .. }.
        """
        copy = self.copy( )
        copy.inplaceSqrt( )
        return copy


    def copy(self , copy_data = True):
        """Will create a deep copy of self, if copy_data is set to False the
        copy will have all z-values set to zero.
        """
        return self._copy( copy_data)


    def write(self , filename):

        """
        Will write the surface as an ascii formatted file to @filename.
        """
        self._write(  filename )



    def assign(self , value):
        """
        Will set all the values in the surface to @value"
        """
        self._assign(value)


    def __setitem__(self , index , value):
        if isinstance(index , int):
            if index >= len(self):
                raise IndexError("Invalid index:%d - valid range [0,%d)" % (index , len(self)))
            if index < 0:
                index += len(self)

            self._iset_zvalue(index , value)
        else:
            raise TypeError("Invalid index type:%s - must be integer" % index)


    def __getitem__(self , index):
        if isinstance(index , int):
            idx = index
            ls = len(self)
            if idx < 0:
                idx += ls
            if 0 <= idx < ls:
                return self._iget_zvalue(idx)
            else:
                raise IndexError("Invalid index:%d - valid range [0,%d)" % (index , len(self)))
        else:
            raise TypeError("Invalid index type:%s - must be integer" % index)


    def getXY(self, index):
        """Gets the index'th (x,y) coordinate"""
        if isinstance(index, int):
            idx = index
            if idx < 0:
                idx += len(self)
            if not 0 <= idx < len(self):
                raise IndexError("Invalid index:%d - valid range [0,%d)" % (index, len(self)))
            index = idx
        else:
            raise TypeError("Invalid index type:%s - must be integer" % index)

        x = ctypes.c_double()
        y = ctypes.c_double()
        self._iget_xy(index, ctypes.byref(x), ctypes.byref(y))

        return x.value, y.value


    def getNX(self):
        return self._get_nx()


    def getNY(self):
        return self._get_ny()

    def getPointset(self):
        return self._get_pointset()

    def _assert_idx_or_i_and_j(self, idx, i, j):
        if idx is None:
            if i is None or j is None:
                raise ValueError('idx is None, i and j must be ints, was %s and %s.' % (i, j))
        else:
            if i is not None or j is not None:
                raise ValueError('idx is set, i and j must be None, was %s and %s.' % (i, j))


    def getXYZ(self, idx=None, i=None, j=None):
        """Returns a tuple of 3 floats, (x,y,z) for given global index, or i and j."""
        self._assert_idx_or_i_and_j(idx, i, j)
        if idx is None:
            nx, ny = self.getNX(), self.getNY()
            i_idx, j_idx = i,j
            if i_idx < 0:
                i_idx += self.getNX()
            if j_idx < 0:
                j_idx += self.getNY()
            if 0 <= i_idx < self.getNX() and 0 <= j_idx < self.getNY():
                idx = j_idx * self.getNX() + i_idx
            else:
                fmt = 'Index error: i=%d not in [0,nx=%d) or j=%d not in [0,ny=%d).'
                raise IndexError(fmt % (i, nx, j, ny))
        x,y = self.getXY(idx)
        z = self[idx]
        return (x,y,z)


    def free(self):
        self._free()

    def __repr__(self):
        cnt = 'nx=%d, ny=%d' % (self.getNX(), self.getNY())
        return self._create_repr(cnt)





from .layer import Layer
from .fault_collection import FaultCollection
from .fault import Fault
from .fault_line import FaultLine
from .fault_segments import FaultSegment , SegmentMap
from .fault_block import FaultBlock , FaultBlockCell
from .fault_block_layer import FaultBlockLayer



from cwrap import BaseCClass

from ecl import EclPrototype
from ecl.well import WellTypeEnum, WellConnection
from ecl.util.util import CTime

class WellState(BaseCClass):
    TYPE_NAME = "well_state"

    _global_connections_size = EclPrototype("int well_conn_collection_get_size(void*)", bind = False)
    _global_connections_iget = EclPrototype("well_connection_ref well_conn_collection_iget(void*, int)", bind = False)
    _segment_collection_size = EclPrototype("int well_segment_collection_get_size(void*)", bind = False)
    _segment_collection_iget = EclPrototype("well_segment_ref well_segment_collection_iget(void*, int)", bind = False)
    _has_global_connections  = EclPrototype("bool  well_state_has_global_connections(well_state)")
    _get_global_connections  = EclPrototype("void* well_state_get_global_connections(well_state)")
    _get_segment_collection  = EclPrototype("void* well_state_get_segments(well_state)")
    _branches                = EclPrototype("void* well_state_get_branches(well_state)")
    _segments                = EclPrototype("void* well_state_get_segments(well_state)")
    _get_name                = EclPrototype("char* well_state_get_name(well_state)")
    _is_open                 = EclPrototype("bool  well_state_is_open(well_state)")
    _is_msw                  = EclPrototype("bool  well_state_is_MSW(well_state)")
    _well_number             = EclPrototype("int   well_state_get_well_nr(well_state)")
    _report_number           = EclPrototype("int   well_state_get_report_nr(well_state)")
    _has_segment_data        = EclPrototype("bool  well_state_has_segment_data(well_state)")
    _sim_time                = EclPrototype("time_t well_state_get_sim_time(well_state)")
    _well_type               = EclPrototype("well_type_enum well_state_get_type(well_state)")
    _oil_rate                = EclPrototype("double well_state_get_oil_rate(well_state)")
    _gas_rate                = EclPrototype("double well_state_get_gas_rate(well_state)")
    _water_rate              = EclPrototype("double well_state_get_water_rate(well_state)")
    _volume_rate             = EclPrototype("double well_state_get_volume_rate(well_state)")
    _oil_rate_si             = EclPrototype("double well_state_get_oil_rate_si(well_state)")
    _gas_rate_si             = EclPrototype("double well_state_get_gas_rate_si(well_state)")
    _water_rate_si           = EclPrototype("double well_state_get_water_rate_si(well_state)")
    _volume_rate_si          = EclPrototype("double well_state_get_volume_rate_si(well_state)")
    _get_global_well_head    = EclPrototype("well_connection_ref well_state_get_global_wellhead(well_state)")
    
    def __init__(self):
        raise NotImplementedError("Class can not be instantiated directly")

    def name(self):
        """ @rtype: str """
        return self._get_name( )

    def isOpen(self):
        """ @rtype: bool """
        return self._is_open( )

    def free(self):
        pass

    def wellHead(self):
        well_head = self._get_global_well_head()
        well_head.setParent( self )
        return well_head

    def wellNumber(self):
        """ @rtype: int """
        return self._well_number( )

    def reportNumber(self):
        """ @rtype: int """
        return self._report_number( )

    def simulationTime(self):
        """ @rtype: CTime """
        return self._sim_time( )

    def wellType(self):
        """ @rtype: WellTypeEnum """
        return self._well_type( )

    def hasGlobalConnections(self):
        """ @rtype: bool """
        return self._has_global_connections( )

    def globalConnections(self):
        """ @rtype: list of WellConnection """
        global_connections = self._get_global_connections( )
        count = self._global_connections_size( global_connections )

        values = []
        for index in range(count):
            value = self._global_connections_iget(global_connections, index).setParent( self )
            values.append(value)
        return values

    def __len__(self):
        return self.numSegments()

    def __getitem__(self, idx):
        return self.igetSegment(idx)

    def numSegments(self):
        """ @rtype: int """
        segment_collection = self._get_segment_collection( )
        count = self._segment_collection_size(segment_collection)
        return count


    def segments(self):
        """ @rtype: list of WellSegment """
        segment_collection = self._get_segment_collection( )

        values = []
        for index in range(self.numSegments()):
            value = self._segment_collection_iget(segment_collection, index).setParent(self)
            values.append(value)

        return values


    def igetSegment(self , seg_idx):
        """ @rtype: WellSegment """
        if seg_idx < 0:
            seg_idx += len(self)

        if not 0 <= seg_idx < self.numSegments():
            raise IndexError("Invalid index:%d - valid range [0,%d)" % (seg_idx , len(self)))

        segment_collection = self._get_segment_collection( )
        return self._segment_collection_iget(segment_collection, seg_idx).setParent(self)

    def isMultiSegmentWell(self):
        """ @rtype: bool """
        return self._is_msw( )

    def hasSegmentData(self):
        """ @rtype: bool """
        return self._has_segment_data( )

    def __repr__(self):
        name = self.name()
        if name:
            name = '%s' % name
        else:
            name = '[no name]'
        msw  = ' (multi segment)' if self.isMultiSegmentWell() else ''
        wn = str(self.wellNumber())
        type_ = self.wellType()
        open_ = 'open' if self.isOpen() else 'shut'
        cnt = '%s%s, number = %s, type = "%s", state = %s' % (name, msw, wn, type_, open_)
        return self._create_repr(cnt)

    def gasRate(self):
        return self._gas_rate( )

    def waterRate(self):
        return self._water_rate( )

    def oilRate(self):
        return self._oil_rate( )

    def volumeRate(self):
        return self._volume_rate( )

    def gasRateSI(self):
        return self._gas_rate_si( )

    def waterRateSI(self):
        return self._water_rate_si( )

    def oilRateSI(self):
        return self._oil_rate_si( )

    def volumeRateSI(self):
        return self._volume_rate_si( )



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_file.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
The ecl_file module contains functionality to load a an ECLIPSE file
in 'restart format'. Files of 'restart format' include restart files,
init files, grid files, summary files and RFT files.

The ecl_file implementation is agnostic[1] to the content and
structure of the file; more specialized classes like EclSum and
EclGrid use the EclFile functionality for low level file loading.

The typical usage involves loading a complete file, and then
subsequently querying for various keywords. In the example below we
load a restart file, and ask for the SWAT keyword:

   file = EclFile( "ECLIPSE.X0067" )
   swat_kw = file.iget_named_kw( "SWAT" , 0 )

The ecl_file module is a thin wrapper around the ecl_file.c
implementation from the libecl library.

[1]: In particular for restart files, which do not have a special
     RestartFile class, there is some specialized functionality.
"""
import re
import types
import datetime
import ctypes

from cwrap import BaseCClass

from ecl import EclPrototype
from ecl.util.util import CTime
from ecl.util.util import monkey_the_camel
from ecl import EclFileEnum
from ecl.eclfile import EclKW, EclFileView


class EclFile(BaseCClass):
    TYPE_NAME = "ecl_file"
    _open                        = EclPrototype("void*       ecl_file_open( char* , int )" , bind = False)
    _get_file_type               = EclPrototype("ecl_file_enum ecl_util_get_file_type( char* , bool* , int*)" , bind = False)
    _writable                    = EclPrototype("bool        ecl_file_writable( ecl_file )")
    _save_kw                     = EclPrototype("void        ecl_file_save_kw( ecl_file , ecl_kw )")
    _close                       = EclPrototype("void        ecl_file_close( ecl_file )")
    _iget_restart_time           = EclPrototype("time_t      ecl_file_iget_restart_sim_date( ecl_file , int )")
    _iget_restart_days           = EclPrototype("double      ecl_file_iget_restart_sim_days( ecl_file , int )")
    _get_restart_index           = EclPrototype("int         ecl_file_get_restart_index( ecl_file , time_t)")
    _get_src_file                = EclPrototype("char*       ecl_file_get_src_file( ecl_file )")
    _replace_kw                  = EclPrototype("void        ecl_file_replace_kw( ecl_file , ecl_kw , ecl_kw , bool)")
    _fwrite                      = EclPrototype("void        ecl_file_fwrite_fortio( ecl_file , fortio , int)")
    _has_report_step             = EclPrototype("bool        ecl_file_has_report_step( ecl_file , int)")
    _has_sim_time                = EclPrototype("bool        ecl_file_has_sim_time( ecl_file , time_t )")
    _get_global_view             = EclPrototype("ecl_file_view_ref ecl_file_get_global_view( ecl_file )")
    _write_index                 = EclPrototype("bool        ecl_file_write_index( ecl_file , char*)")
    _fast_open                   = EclPrototype("void*       ecl_file_fast_open( char* , char* , int )" , bind=False)


    @staticmethod
    def get_filetype(filename):
        fmt_file    = ctypes.c_bool()
        report_step = ctypes.c_int()

        file_type = EclFile._get_file_type( filename , ctypes.byref( fmt_file ) , ctypes.byref(report_step))
        if file_type in [EclFileEnum.ECL_RESTART_FILE , EclFileEnum.ECL_SUMMARY_FILE]:
            report_step = report_step.value
        else:
            report_step = None

        if file_type in [EclFileEnum.ECL_OTHER_FILE , EclFileEnum.ECL_DATA_FILE]:
            fmt_file = None
        else:
            fmt_file = fmt_file.value


        return (file_type , report_step , fmt_file )



    @classmethod
    def restart_block( cls , filename , dtime = None , report_step = None):
        raise NotImplementedError("The restart_block implementation has been removed - open file normally and use EclFileView.")



    @classmethod
    def contains_report_step( cls , filename , report_step ):
        """
        Will check if the @filename contains @report_step.

        This classmethod works by opening the file @filename and
        searching through linearly to see if an ecl_kw with value
        corresponding to @report_step can be found. Since this is a
        classmethod it is invoked like this:

           import ecl.ecl.ecl as ecl
           ....
           if ecl.EclFile.contains_report_step("ECLIPSE.UNRST" , 20):
              print "OK - file contains report step 20"
           else:
              print "File does not contain report step 20"

        If you have already loaded the file into an EclFile instance
        you should use the has_report_step() method instead.
        """
        obj = EclFile( filename )
        return obj.has_report_step( report_step )


    @classmethod
    def contains_sim_time( cls , filename , dtime ):
        """
        Will check if the @filename contains simulation at @dtime.

        This classmethod works by opening the file @filename and
        searching through linearly to see if a result block at the
        time corresponding to @dtime can be found. Since this is a
        classmethod it is invoked like this:

           import ecl.ecl.ecl as ecl
           ....
           if ecl.EclFile.contains_sim_time("ECLIPSE.UNRST" , datetime.datetime( 2007 , 10 , 10) ):
              print "OK - file contains 10th of October 2007"
           else:
              print "File does not contain 10th of October 2007"

        If you have already loaded the file into an EclFile instance
        you should use the has_sim_time() method instead.
        """
        obj = EclFile( filename )
        return obj.has_sim_time( dtime )

    @property
    def report_list(self):
        report_steps = []
        try:
            seqnum_list = self["SEQNUM"]
            for s in seqnum_list:
                report_steps.append( s[0] )
        except KeyError:
            # OK - we did not have seqnum; that might be because this
            # a non-unified restart file; or because this is not a
            # restart file at all.
            fname = self.getFilename( )
            matchObj = re.search("\.[XF](\d{4})$" , fname)
            if matchObj:
                report_steps.append( int(matchObj.group(1)) )
            else:
                raise TypeError('Tried get list of report steps from file "%s" - which is not a restart file' % fname)


        return report_steps



    @classmethod
    def file_report_list( cls , filename ):
        """
        Will identify the available report_steps from @filename.
        """

        file = EclFile( filename )
        return file.report_list



    def __repr__(self):
        fn = self.getFilename()
        wr = ', read/write' if self._writable() else ''
        return self._create_repr('"%s"%s' % (fn,wr))


    def __init__( self , filename , flags = 0 , index_filename = None):
        """
        Loads the complete file @filename.

        Will create a new EclFile instance with the content of file
        @filename. The file @filename must be in 'restart format' -
        otherwise it will be crash and burn.

        The optional argument flags can be an or'ed combination of the
        flags:

           ecl.ECL_FILE_WRITABLE : It is possible to update the
              content of the keywords in the file.

           ecl.ECL_FILE_CLOSE_STREAM : The underlying FILE * is closed
              when not used; to save number of open file descriptors
              in cases where a high number of EclFile instances are
              open concurrently.

        When the file has been loaded the EclFile instance can be used
        to query for and get reference to the EclKW instances
        constituting the file, like e.g. SWAT from a restart file or
        FIPNUM from an INIT file.
        """
        if index_filename is None:
            c_ptr = self._open( filename , flags )
        else:
            c_ptr = self._fast_open(filename, index_filename, flags)

        if c_ptr is None:
            raise IOError('Failed to open file "%s"' % filename)
        else:
            super(EclFile , self).__init__(c_ptr)
            self.global_view = self._get_global_view( )
            self.global_view.setParent( self )


    def save_kw( self , kw ):
        """
        Will write the @kw back to file.

        This function should typically be used in situations like this:

          1. Create an EclFile instance around an ECLIPSE output file.
          2. Extract a keyword of interest and modify it.
          3. Call this method to save the modifications to disk.

        There are several restrictions to the use of this function:

          1. The EclFile instance must have been created with the
             optional read_only flag set to False.

          2. You can only modify the content of the keyword; if you
             try to modify the header in any way (i.e. size, datatype
             or name) the function will fail.

          3. The keyword you are trying to save must be exactly the
             keyword you got from this EclFile instance, otherwise the
             function will fail.
        """
        if self._writable( ):
            self._save_kw(  kw )
        else:
            raise IOError('save_kw: the file "%s" has been opened read only.' % self.getFilename( ))


    def __len__(self):
        return len(self.global_view)


    def close(self):
        if self:
            self._close( )
            self._invalidateCPointer( )


    def free(self):
        self.close()


    def block_view(self, kw, kw_index):
        if not kw in self:
            raise KeyError('No such keyword "%s".' % kw)
        ls = self.global_view.numKeywords(kw)
        idx = kw_index
        if idx < 0:
            idx += ls
        if 0 <= idx < ls:
            return self.global_view.blockView(kw, idx)
        raise IndexError('Index out of range, must be in [0, %d), was %d.' % (ls, kw_index))


    def block_view2(self, start_kw, stop_kw, start_index):
        return self.global_view.blockView2( start_kw , stop_kw, start_index )


    def restart_view(self, seqnum_index=None, report_step=None, sim_time=None, sim_days=None):
        return self.global_view.restartView( seqnum_index, report_step , sim_time, sim_days )


    def select_block(self, kw, kw_index):
        raise NotImplementedError("The select_block implementation has been removed - use EclFileView")


    def select_global( self ):
        raise NotImplementedError("The select_global implementation has been removed - use EclFileView")


    def select_restart_section( self, index = None , report_step = None , sim_time = None):
        raise NotImplementedError("The select_restart_section implementation has been removed - use EclFileView")
        """
        Will select a restart section as the active section.

        You must specify a report step with the @report_step argument,
        a true time with the @sim_time argument or a plain index to
        select restart block. If none of arguments are given exception
        TypeError will be raised. If present the @sim_time argument
        should be a datetime instance.

        If the restart section you ask for can not be found the method
        will raise a ValueError exeception. To protect against this
        you can query first with the has_report_step(),
        has_sim_time() or num_report_steps() methods.

        This method should be used when you have already loaded the
        complete file; if you only want to load a section from the
        file you can use the classmethod restart_block().

        The method will return 'self' which can be used to aid
        readability.
        """



    def select_last_restart( self ):
        raise NotImplementedError("The select_restart_section implementation has been removed - use EclFileView")
        """
        Will select the last SEQNUM block in restart file.

        Works by searching for the last SEQNUM keyword; the SEQNUM
        Keywords are only present in unified restart files. If this
        is a non-unified restart file (or not a restart file at all),
        the method will do nothing and return False.
        """





    def __getitem__(self , index):
        """
        Implements [] operator; index can be integer or key.

        Will look up EclKW instances from the current EclFile
        instance. The @index argument can either be an integer, in
        which case the method will return EclKW number @index, or
        alternatively a keyword string, in which case the method will
        return a list of EclKW instances with that keyword:

           restart_file = ecl_file.EclFile("ECLIPSE.UNRST")
           kw9 = restart_file[9]
           swat_list = restart_file["SWAT"]

        The keyword based lookup can be combined with an extra [] to
        get EclKW instance nr:

           swat9 = restart_file["SWAT"][9]

        Will return the 10'th SWAT keyword from the restart file. The
        following example will iterate over all the SWAT keywords in a
        restart file:

           restart_file = ecl_file.EclFile("ECLIPSE.UNRST")
           for swat in restart_file["SWAT"]:
               ....
        """
        if isinstance(index, int):
            ls = len(self)
            idx = index
            if idx < 0:
                idx += ls
            if 0 <= idx < ls:
                return self.global_view[idx]
            else:
                raise IndexError('Index must be in [0, %d), was: %d.' % (ls, index))
        return self.global_view[index]


    def iget_kw( self , index , copy = False):
        """
        Will return EclKW instance nr @index.

        In the files loaded with the EclFile implementation the
        ECLIPSE keywords come sequentially in a long series, an INIT
        file might have the following keywords:

          INTEHEAD
          LOGIHEAD
          DOUBHEAD
          PORV
          DX
          DY
          DZ
          PERMX
          PERMY
          PERMZ
          MULTX
          MULTY
          .....

        The iget_kw() method will give you a EclKW reference to
        keyword nr @index. This functionality is also available
        through the index operator []:

           file = EclFile( "ECLIPSE.INIT" )
           permx = file.iget_kw( 7 )
           permz = file[ 9 ]

        Observe that the returned EclKW instance is only a reference
        to the data owned by the EclFile instance.

        The method iget_named_kw() which lets you specify the name of
        the keyword you are interested in is in general more useful
        than this method.
        """
        kw = self[index]
        if copy:
            return EclKW.copy( kw )
        else:
            return kw


    def iget_named_kw( self , kw_name , index , copy = False):
        return self.global_view.iget_named_kw( kw_name , index )



    def restart_get_kw( self , kw_name , dtime , copy = False):
        """Will return EclKW @kw_name from restart file at time @dtime.

        This function assumes that the current EclFile instance
        represents a restart file. It will then look for keyword
        @kw_name exactly at the time @dtime; @dtime is a datetime
        instance:

            file = EclFile( "ECLIPSE.UNRST" )
            swat2010 = file.restart_get_kw( "SWAT" , datetime.datetime( 2000 , 1 , 1 ))

        By default the returned kw instance is a reference to the
        ecl_kw still contained in the EclFile instance; i.e. the kw
        will become a dangling reference if the EclFile instance goes
        out of scope. If the optional argument @copy is True the
        returned kw will be a true copy.

        If the file does not have the keyword at the specified time
        the function will raise IndexError(); if the file does not
        have the keyword at all - KeyError will be raised.
        """
        index = self._get_restart_index( CTime( dtime ) )
        if index >= 0:
            if self.num_named_kw(kw_name) > index:
                kw = self.iget_named_kw( kw_name , index )
                if copy:
                    return EclKW.copy( kw )
                else:
                    return kw
            else:
                if self.has_kw(kw_name):
                    raise IndexError('Does not have keyword "%s" at time:%s.' % (kw_name , dtime))
                else:
                    raise KeyError('Keyword "%s" not recognized.' % kw_name)
        else:
            raise IndexError('Does not have keyword "%s" at time:%s.' % (kw_name , dtime))


    def replace_kw( self , old_kw , new_kw):
        """
        Will replace @old_kw with @new_kw in current EclFile instance.

        This method can be used to replace one of the EclKW instances
        in the current EclFile. The @old_kw reference must be to the
        actual EclKW instance in the current EclFile instance (the
        final comparison is based on C pointer equality!), i.e. it
        must be a reference (not a copy) from one of the ??get_kw??
        methods of the EclFile class. In the example below we replace
        the SWAT keyword from a restart file:

           swat = file.iget_named_kw( "SWAT" , 0 )
           new_swat = swat * 0.25
           file.replace_kw( swat , new_swat )


        The C-level ecl_file_type structure takes full ownership of
        all installed ecl_kw instances; mixing the garbage collector
        into it means that this is quite low level - and potentially
        dangerous!
        """

        # We ensure that this scope owns the new_kw instance; the
        # new_kw will be handed over to the ecl_file instance, and we
        # can not give away something we do not alreeady own.
        if not new_kw.data_owner:
            new_kw = EclKW.copy( new_kw )

        # The ecl_file instance will take responsability for freeing
        # this ecl_kw instance.
        new_kw.data_owner = False
        self._replace_kw( old_kw , new_kw , False )



    @property
    def size(self):
        """
        The number of keywords in the current EclFile object.
        """
        return len(self)

    @property
    def unique_size( self ):
        """
        The number of unique keyword (names) in the current EclFile object.
        """
        return self.global_view.uniqueSize( )



    def keys(self):
        """
        Will return a list of unique kw names - like keys() on a dict.
        """
        header_dict = {}
        for index in range(len(self)):
            kw = self[index]
            header_dict[ kw.getName() ] = True
        return header_dict.keys()


    @property
    def headers(self):
        """
        Will return a list of the headers of all the keywords.
        """
        header_list = []
        for index in range(self.size):
            kw = self[index]
            header_list.append( kw.header )
        return header_list

    @property
    def report_steps( self ):
        """
        Will return a list of all report steps.

        The method works by iterating through the whole restart file
        looking for 'SEQNUM' keywords; if the current EclFile instance
        is not a restart file it will not contain any 'SEQNUM'
        keywords and the method will simply return an empty list.
        """
        steps = []
        seqnum_list = self["SEQNUM"]
        for kw in self["SEQNUM"]:
            steps.append( kw[0] )
        return steps

    @property
    def report_dates( self ):
        """
        Will return a list of the dates for all report steps.

        The method works by iterating through the whole restart file
        looking for 'SEQNUM/INTEHEAD' keywords; the method can
        probably be tricked by other file types also containing an
        INTEHEAD keyword.
        """
        if self.has_kw('SEQNUM'):
            dates = []
            for index in range( self.num_named_kw( 'SEQNUM' )):
                dates.append( self.iget_restart_sim_time( index ))
            return dates
        elif 'INTEHEAD' in self:
            # This is a uber-hack; should export the ecl_rsthead
            # object as ctypes structure.
            intehead = self["INTEHEAD"][0]
            year = intehead[66]
            month = intehead[65]
            day = intehead[64]
            date = datetime.datetime( year , month , day )
            return [ date ]
        return None


    @property
    def dates( self ):
        """
        Will return a list of the dates for all report steps.
        """
        return self.report_dates


    def num_named_kw( self , kw):
        """
        The number of keywords with name == @kw in the current EclFile object.
        """
        return self.global_view.numKeywords( kw )


    def has_kw( self , kw , num = 0):
        """
        Check if current EclFile instance has a keyword @kw.

        If the optional argument @num is given it will check if the
        EclFile has at least @num occurences of @kw.
        """

        return self.num_named_kw( kw ) > num

    def __contains__(self , kw):
        """
        Check if the current file contains keyword @kw.
        """
        return self.has_kw( kw )

    def has_report_step( self , report_step ):
        """
        Checks if the current EclFile has report step @report_step.

        If the EclFile in question is not a restart file, you will
        just get False. If you want to check if the file contains the
        actual report_step before loading the file, you should use the
        classmethod contains_report_step() instead.
        """
        return self._has_report_step( report_step )

    def num_report_steps( self ):
        """
        Returns the total number of report steps in the restart file.

        Works by counting the number of 'SEQNUM' instances, and will
        happily return 0 for a non-restart file. Observe that the
        report_steps present in a unified restart file are in general
        not consecutive, i.e. the last report step will typically be
        much higher than the return value from this function.
        """
        return len( self["SEQNUM"] )



    def has_sim_time( self , dtime ):
        """
        Checks if the current EclFile has data for time @dtime.

        The implementation goes through all the INTEHEAD headers in
        the EclFile, i.e. it can be fooled (and probably crash and
        burn) if the EclFile instance in question is has INTEHEAD
        keyword(s), but is still not a restart file. The @dtime
        argument should be a normal python datetime instance.
        """
        return self._has_sim_time( CTime(dtime) )


    def iget_restart_sim_time( self , index ):
        """
        Will locate restart block nr @index and return the true time
        as a datetime instance.
        """
        ct = CTime(self._iget_restart_time( index ))
        return ct.datetime()


    def iget_restart_sim_days( self , index ):
        """
        Will locate restart block nr @index and return the number of days
        (in METRIC at least ...) since the simulation started.

        """
        return self._iget_restart_days( index )


    def get_filename(self):
        """
        Name of the file currently loaded.
        """
        fn = self._get_src_file()
        return str(fn) if fn else ''

    def fwrite( self , fortio ):
        """
        Will write current EclFile instance to fortio stream.

        ECLIPSE is written in Fortran; and a "special" handle for
        Fortran IO must be used when reading and writing these files.
        This method will write the current EclFile instance to a
        FortIO stream already opened for writing:

           import ecl.ecl.ecl as ecl
           ...
           fortio = ecl.FortIO( "FILE.XX" )
           file.fwrite( fortio )
           fortio.close()

        """
        self._fwrite(  fortio , 0 )

    def write_index(self, index_file_name):
        if not self._write_index(index_file_name):
            raise IOError("Failed to write index file:%s" % index_file_name)


class EclFileContextManager(object):

    def __init__(self , ecl_file):
        self.__ecl_file = ecl_file

    def __enter__(self):
        return self.__ecl_file

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__ecl_file.close()
        return False


def openEclFile( file_name , flags = 0):
    print('The function openEclFile is deprecated, use open_ecl_file.')
    return open_ecl_file(file_name, flags)

def open_ecl_file(file_name, flags=0):
    return EclFileContextManager(EclFile(file_name, flags))



monkey_the_camel(EclFile, 'getFileType', EclFile.get_filetype, staticmethod)
monkey_the_camel(EclFile, 'blockView', EclFile.block_view)
monkey_the_camel(EclFile, 'blockView2', EclFile.block_view2)
monkey_the_camel(EclFile, 'restartView', EclFile.restart_view)
monkey_the_camel(EclFile, 'getFilename', EclFile.get_filename)



#  Copyright (C) 2013  Statoil ASA, Norway. 
#   
#  The file 'test_run.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, teither version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 
import random
import os.path
import subprocess
import argparse
from  .test_area import TestAreaContext


def path_exists( path ):
    if os.path.exists( path ):
        return (True , "Path:%s exists" % path)
    else:
        return (False , "ERROR: Path:%s does not exist" % path)


class TestRun(object):
    default_ert_cmd = "ert"
    default_ert_version = "stable"
    default_path_prefix = None

    def __init__(self , config_file , args = [] , name = None):
        if os.path.exists( config_file ) and os.path.isfile( config_file ):
            self.parseArgs(args)
            self.__ert_cmd = TestRun.default_ert_cmd
            self.path_prefix = TestRun.default_path_prefix
            self.config_file = config_file
            
            self.check_list = []
            self.workflows = []
            if name:
                self.name = name
            else:
                self.name = config_file.replace("/" , ".")
                while True:
                    if self.name[0] == ".":
                        self.name = self.name[1:]
                    else:
                        break
            self.name += "/%08d" % random.randint(0,100000000)
        else:
            raise IOError("No such config file: %s" % config_file)

    
    def parseArgs(self , args):
        parser = argparse.ArgumentParser()
        parser.add_argument("-v" , "--version" , default = self.default_ert_version)
        parser.add_argument("args" , nargs="*")
        result = parser.parse_args(args)
        self.ert_version = result.version
        self.args = result.args
        

    def get_config_file(self):
        return self.__config_file

    def set_config_file(self , input_config_file):
        self.__config_file = os.path.basename( input_config_file )
        self.abs_config_file = os.path.abspath( input_config_file )

    config_file = property( get_config_file , set_config_file )

    #-----------------------------------------------------------------

    def set_path_prefix(self , path_prefix):
        self.__path_prefix = path_prefix
    
    def get_path_prefix(self):
        return self.__path_prefix

    path_prefix = property( get_path_prefix , set_path_prefix )

    #-----------------------------------------------------------------
    
    def get_ert_cmd(self):
        return self.__ert_cmd

    def set_ert_cmd(self , cmd):
        self.__ert_cmd = cmd

    ert_cmd = property( get_ert_cmd , set_ert_cmd)

    #-----------------------------------------------------------------

    def get_workflows(self):
        return self.workflows

    
    def add_workflow(self , workflow):
        self.workflows.append( workflow )

    #-----------------------------------------------------------------

    def get_args(self):
        return self.args

    #-----------------------------------------------------------------

    def add_check( self , check_func , arg):
        if callable(check_func):
            self.check_list.append( (check_func , arg) )
        else:
            raise Exception("The checker:%s is not callable" % check_func )

    #-----------------------------------------------------------------

    def __run(self , work_area ):
        argList = [ self.ert_cmd , "-v" , self.ert_version ]
        for arg in self.args:
            argList.append( arg )

        argList.append( self.config_file )
        for wf in self.workflows:
            argList.append( wf )

        status = subprocess.call( argList )
        if status == 0:
            return (True , "ert has run successfully")
        else:
            return (False , "ERROR:: ert exited with status code:%s" % status)

    
    def run(self):
        if len(self.workflows):
            with TestAreaContext(self.name , prefix = self.path_prefix , store_area = False) as work_area:
                test_cwd = work_area.get_cwd()
                work_area.copy_parent_content( self.abs_config_file )
                status = self.__run( work_area )
                global_status = status[0]

                status_list = [ status ]
                if status[0]:
                    for (check_func , arg) in self.check_list:
                        status = check_func( arg )
                        status_list.append( status )
                        if not status[0]:
                            global_status = False

                if not global_status:
                    work_area.set_store( True )
                    
            return (global_status , test_cwd , status_list)
        else:
            raise Exception("Must have added workflows before invoking start()")
            



#  Copyright (C) 2017  Statoil ASA, Norway.
#
#  The file 'ecl_type.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCClass, BaseCEnum
from ecl import EclPrototype

class EclTypeEnum(BaseCEnum):
    TYPE_NAME="ecl_type_enum"
    ECL_CHAR_TYPE   = None
    ECL_FLOAT_TYPE  = None
    ECL_DOUBLE_TYPE = None
    ECL_INT_TYPE    = None
    ECL_BOOL_TYPE   = None
    ECL_MESS_TYPE   = None
    ECL_STRING_TYPE = None

EclTypeEnum.addEnum("ECL_CHAR_TYPE",   0)
EclTypeEnum.addEnum("ECL_FLOAT_TYPE",  1)
EclTypeEnum.addEnum("ECL_DOUBLE_TYPE", 2)
EclTypeEnum.addEnum("ECL_INT_TYPE",    3)
EclTypeEnum.addEnum("ECL_BOOL_TYPE",   4)
EclTypeEnum.addEnum("ECL_MESS_TYPE",   5)
EclTypeEnum.addEnum("ECL_STRING_TYPE", 7)

#-----------------------------------------------------------------

class EclDataType(BaseCClass):

    TYPE_NAME = "ecl_data_type"

    _alloc            = EclPrototype("void* ecl_type_alloc_python(ecl_type_enum, size_t)", bind=False)
    _alloc_from_type  = EclPrototype("void* ecl_type_alloc_from_type_python(ecl_type_enum)", bind=False)
    _alloc_from_name  = EclPrototype("void* ecl_type_alloc_from_name_python(char*)", bind=False)
    _free             = EclPrototype("void ecl_type_free_python(ecl_data_type)")
    _get_type         = EclPrototype("ecl_type_enum ecl_type_get_type_python(ecl_data_type)")
    _get_element_size = EclPrototype("size_t ecl_type_get_sizeof_iotype_python(ecl_data_type)")
    _is_int           = EclPrototype("bool ecl_type_is_int_python(ecl_data_type)")
    _is_char          = EclPrototype("bool ecl_type_is_char_python(ecl_data_type)")
    _is_float         = EclPrototype("bool ecl_type_is_float_python(ecl_data_type)")
    _is_double        = EclPrototype("bool ecl_type_is_double_python(ecl_data_type)")
    _is_mess          = EclPrototype("bool ecl_type_is_mess_python(ecl_data_type)")
    _is_bool          = EclPrototype("bool ecl_type_is_bool_python(ecl_data_type)")
    _is_string        = EclPrototype("bool ecl_type_is_string_python(ecl_data_type)")
    _get_name         = EclPrototype("char* ecl_type_alloc_name_python(ecl_data_type)")
    _is_numeric       = EclPrototype("bool ecl_type_is_numeric_python(ecl_data_type)")
    _is_equal         = EclPrototype("bool ecl_type_is_equal_python(ecl_data_type, ecl_data_type)")

    def __init__(self, type_enum=None, element_size=None, type_name=None):
        self._assert_valid_arguments(type_enum, element_size, type_name)

        if type_name:
            c_ptr = self._alloc_from_name(type_name)
        elif element_size is None:
            c_ptr = self._alloc_from_type(type_enum)
        else:
            c_ptr = self._alloc(type_enum, element_size)

        super(EclDataType, self).__init__(c_ptr)

    def _assert_valid_arguments(self, type_enum, element_size, type_name):
        if type_name is not None:
            if type_enum is not None or element_size is not None:
                err_msg = ("Type name given (%s). Expected both " +
                        "type_enum and element_size to be None")
                raise ValueError(err_msg % type_name)

        elif type_enum is None:
            raise ValueError("Both type_enum and type_name is None!")

        elif type_enum == EclTypeEnum.ECL_STRING_TYPE:
            if element_size is None:
                raise ValueError("When creating an ECL_STRING one must " +
                        "provide an element size!")

            if not (0 <= element_size <= 999):
                raise ValueError("Expected element_size to be in the range " +
                        "[0, 999], was: %d" % element_size)

    @property
    def type(self):
        return self._get_type()

    @property
    def element_size(self):
        return self._get_element_size()

    @property
    def type_name(self):
        return self._get_name()

    def free(self):
        self._free()

    def is_int(self):
        return self._is_int()

    def is_char(self):
        return self._is_char()

    def is_float(self):
        return self._is_float()

    def is_double(self):
        return self._is_double()

    def is_mess(self):
        return self._is_mess()

    def is_bool(self):
        return self._is_bool()

    def is_string(self):
        return self._is_string()

    def is_numeric(self):
        return self._is_numeric()

    def is_equal(self, other):
        return self._is_equal(other)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.is_equal(other)
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.type, self. element_size))

    @classmethod
    def create_from_type_name(cls, name):
        return EclDataType(type_name=name)

    # Enables one to fetch a type as EclDataType.ECL_XXXX
    class classproperty(object):

        def __init__(self, fget):
            self.fget = fget

        def __get__(self, owner_self, owner_cls):
            return self.fget(owner_cls)

    @classproperty
    def ECL_INT(cls):
        return EclDataType(EclTypeEnum.ECL_INT_TYPE)

    @classproperty
    def ECL_FLOAT(cls):
        return EclDataType(EclTypeEnum.ECL_FLOAT_TYPE)

    @classproperty
    def ECL_DOUBLE(cls):
        return EclDataType(EclTypeEnum.ECL_DOUBLE_TYPE)

    @classproperty
    def ECL_BOOL(cls):
        return EclDataType(EclTypeEnum.ECL_BOOL_TYPE)

    @classproperty
    def ECL_MESS(cls):
        return EclDataType(EclTypeEnum.ECL_MESS_TYPE)

    @classproperty
    def ECL_CHAR(cls):
        return EclDataType(EclTypeEnum.ECL_CHAR_TYPE)

    @classmethod
    def ECL_STRING(cls, elem_size):
        return EclDataType(EclTypeEnum.ECL_STRING_TYPE, elem_size)



#  Copyright (C) 2011  Statoil ASA, Norway. 
#   
#  The file 'ctime.py' is part of ERT - Ensemble based Reservoir Tool. 
#   
#  ERT is free software: you can redistribute it and/or modify 
#  it under the terms of the GNU General Public License as published by 
#  the Free Software Foundation, either version 3 of the License, or 
#  (at your option) any later version. 
#   
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY 
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or 
#  FITNESS FOR A PARTICULAR PURPOSE.   
#   
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html> 
#  for more details. 


import ctypes
import datetime
import time

from cwrap import BaseCValue
from ecl import EclPrototype


class CTime(BaseCValue):
    TYPE_NAME = "time_t"
    DATA_TYPE = ctypes.c_long
    _timezone = EclPrototype("char* util_get_timezone()" , bind = False)
    _timegm = EclPrototype("long util_make_datetime_utc(int, int, int, int, int, int)" , bind = False)

    def __init__(self, value):
        if isinstance(value, int):
            value = value
        elif isinstance(value, CTime):
            value = value.value()
        elif isinstance(value, datetime.datetime):
            value = CTime._timegm(value.second, value.minute, value.hour, value.day, value.month, value.year)
        elif isinstance(value, datetime.date):
            value = CTime._timegm(0, 0, 0, value.day, value.month, value.year)
        else:
            raise NotImplementedError("Can not convert class %s to CTime" % value.__class__)

        super(CTime, self).__init__(value)

    def ctime(self):
        """ @rtype: int """
        return self.value()

    def time(self):
        """Return this time_t as a time.gmtime() object"""
        return time.gmtime(self.value())

    def date(self):
        """Return this time_t as a datetime.date([year, month, day])"""
        return datetime.date(*self.time()[0:3])

    def datetime(self):
        return datetime.datetime(*self.time()[0:6])

    def __str__(self):
        return self.datetime().strftime("%Y-%m-%d %H:%M:%S%z")

    def __ge__(self, other):
        return self > other or self == other

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        if isinstance(other, CTime):
            return self.value() > other.value()
        elif isinstance(other, (int, datetime.datetime, datetime.date)):
            return self > CTime(other)
        else:
            raise TypeError("CTime does not support type: %s" % other.__class__)

    def __lt__(self, other):
        if isinstance(other, CTime):
            return self.value() < other.value()
        elif isinstance(other, (int, datetime.datetime, datetime.date)):
            return self < CTime(other)
        else:
            raise TypeError("CTime does not support type: %s" % other.__class__)

    def __ne__(self, other):
        return not self == other

    def __eq__(self, other):
        if isinstance(other, CTime):
            return self.value() == other.value()
        elif isinstance(other, (int, datetime.datetime, datetime.date)):
            return self == CTime(other)
        elif isinstance(other, type(None)):
            return False
        else:
            raise TypeError("CTime does not support type: %s" % other.__class__)

    def __imul__(self, other):
        value = int(self.value() * other)
        self.setValue(value)
        return self

    def __hash__(self):
        return hash(self.value())

    def __iadd__(self, other):
        if isinstance(other, CTime):
            self.setValue(self.value() + other.value())
            return self
        else:
            self.setValue(self.value() + CTime(other).value())
            return self

    def __add__(self, other):
        copy = CTime(self)
        copy += other
        return copy

    def __radd__(self, other):
        return self + other

    def __mul__(self, other):
        copy = CTime(self)
        copy *= other
        return copy

    def __rmul__(self, other):
        return self * other

    def timetuple(self):
        # this function is a requirement for comparing against datetime objects where the CTime is on the right side
        pass

    def __repr__(self):
        return "time_t value: %d [%s]" % (self.value(), str(self))

    @property
    def stripped(self):
        return time.strptime(self, "%Y-%m-%d %H:%M:S%")

    @classmethod
    def timezone(cls):
        """
         Returns the current timezone "in" C
         @rtype: str
        """
        return CTime._timezone()



from math import sqrt
import sys
import six

class GeometryTools(object):
    EPSILON = 0.000001

    @staticmethod
    def lineIntersection(p1, p2, p3, p4):
        """
        Finds intersection between line segments. Returns None if no intersection found.
        Algorithm provided by Paul Bourke

        @type p1: tuple of (float, float)
        @type p2: tuple of (float, float)
        @type p3: tuple of (float, float)
        @type p4: tuple of (float, float)
        @rtype: tuple of (float, float)
        """

        denominator = (p4[1] - p3[1]) * (p2[0] - p1[0]) - (p4[0] - p3[0]) * (p2[1] - p1[1])
        numerator_a = (p4[0] - p3[0]) * (p1[1] - p3[1]) - (p4[1] - p3[1]) * (p1[0] - p3[0])
        numerator_b = (p2[0] - p1[0]) * (p1[1] - p3[1]) - (p2[1] - p1[1]) * (p1[0] - p3[0])

        # coincident?
        if abs(numerator_a) < GeometryTools.EPSILON and abs(numerator_b) < GeometryTools.EPSILON and abs(denominator) < GeometryTools.EPSILON:
            return None

        # parallel?
        if abs(denominator) < GeometryTools.EPSILON:
            return None


        # intersection along the segments?
        mua = numerator_a / denominator
        mub = numerator_b / denominator

        if mua < 0.0 or mua > 1.0 or mub < 0.0 or mub > 1.0:
            return None

        x = p1[0] + mua * (p2[0] - p1[0])
        y = p1[1] + mua * (p2[1] - p1[1])
        return x, y

    @staticmethod
    def polylinesIntersect(polyline1 , polyline2):
        """Test if the polylines polyline1 and polyline2 intersect.
        
        The input arguments must be either Polyline instances[1], or a
        list of (float,float) tuples. The method performs a super
        naive n^2 check and should not be used for large polyline objects.

        @type polyline1: Polyline or list of tuple of (float, float)
        @type polyline2: Polyline or list of tuple of (float, float)

        [1]: The z - coordinate will be ignored.

        """

        for index1 in range(len(polyline1) - 1):
            p1 = polyline1[index1]
            p2 = polyline1[index1 + 1]
            for index2 in range(len(polyline2) - 1):
                p3 = polyline2[index2]
                p4 = polyline2[index2 + 1]

                if GeometryTools.lineIntersection(p1,p2,p3,p4):
                    return True
        
        return False




    @staticmethod
    def ccw(p1, p2, p3):
        """
        Three points are a counter-clockwise turn if ccw > 0, clockwise if
        ccw < 0, and collinear if ccw = 0 because ccw is a determinant that
        gives the signed area of the triangle formed by p1, p2 and p3.

        @type p1: tuple of (float, float)
        @type p2: tuple of (float, float)
        @type p3: tuple of (float, float)
        @rtype: float
        """
        return (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0])


    @staticmethod
    def convexHull(points):
        """
        Given a list of points finds the convex hull
        @type points: list of tuple of (float, float)
        @rtype: list of tuple of (float, float)
        """
        points = sorted(points)

        def keepLeft(hull, r):
            while len(hull) > 1 and GeometryTools.ccw(hull[-2], hull[-1], r) > 0:
                hull.pop()

            if len(hull) == 0 or hull[-1] != r:
                hull.append(r)

            return hull

        l = six.functools.reduce(keepLeft, points, [])
        u = six.functools.reduce(keepLeft, reversed(points), [])
        l.extend([u[i] for i in six.moves.xrange(1, len(u) - 1)])

        return l


    @staticmethod
    def pointInPolygon(p, polygon):
        """
        Finds out if a point is inside a polygon or not
        @type p: tuple of (float, float)
        @type polygon: Polyline or list of tuple of (float, float)
        @rtype: bool
        """
        x = p[0]
        y = p[1]
        n = len(polygon)

        inside = False

        p1x, p1y = polygon[0][0:2]
        for index in range(n + 1):
            p2x, p2y = polygon[index % n][0:2]

            if min(p1y, p2y) < y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x

                    if p1x == p2x or x <= xints:
                        inside = not inside

            p1x, p1y = p2x, p2y

        return inside


    @staticmethod
    def extendToEdge(bounding_polygon, poly_line):
        """
        """
        assert(bounding_polygon.isClosed())
        for p in poly_line:
            if not GeometryTools.pointInPolygon( p , bounding_polygon):
                raise ValueError("The point:%s was not inside bounding polygon")
        
        p1 = poly_line[0]
        ray1 = GeometryTools.lineToRay(poly_line[1], poly_line[0])
        intersection1 = GeometryTools.rayPolygonIntersections(p1, ray1, bounding_polygon)[0] # assume convex
        
        
        p2 = poly_line[-1]
        assert(GeometryTools.pointInPolygon(p2 , bounding_polygon))
                
        ray2 = GeometryTools.lineToRay(poly_line[-2], poly_line[-1])
        intersection2 = GeometryTools.rayPolygonIntersections(p2, ray2, bounding_polygon)
        intersection2 = GeometryTools.rayPolygonIntersections(p2, ray2, bounding_polygon)[0] # assume convex

        return [intersection1[1]] + poly_line + [intersection2[1]]


    @staticmethod
    def slicePolygon(bounding_polygon, poly_line):
        """
        This algorithm extends the end-points of the line and find intersections between the line
        and the enclosing polygon. The result is a polygon sliced by the extended line.

        The enclosing polygon must be convex, closed and completely enclose the line.

        @type bounding_polygon: Polyline or list of tuple of (float, float)
        @type poly_line:  Polyline or list of tuple of (float, float)
        @rtype: list of tuple of (float, float)
        """

        p1 = poly_line[0]
        ray1 = GeometryTools.lineToRay(poly_line[1], poly_line[0])
        tmp = GeometryTools.rayPolygonIntersections(p1, ray1, bounding_polygon)
        intersection1 = GeometryTools.rayPolygonIntersections(p1, ray1, bounding_polygon)[0] # assume convex

        p2 = poly_line[-1]
        ray2 = GeometryTools.lineToRay(poly_line[-2], poly_line[-1])
        intersection2 = GeometryTools.rayPolygonIntersections(p2, ray2, bounding_polygon)[0] # assume convex


        # Check for intersection between the polyline extensions on the inside of the bounadary
        internal_intersection = GeometryTools.lineIntersection( p1 , intersection1[1] , p2 , intersection2[1])
        if internal_intersection:
            start_point = poly_line[0]
            return poly_line + [ internal_intersection , start_point]



        if intersection2[0] < intersection1[0]:
            intersection1, intersection2 = intersection2, intersection1
            poly_line = list(reversed(poly_line))

        result = [intersection1[1]]

        for index in range(intersection1[0] + 1, intersection2[0] + 1):
            result.append(bounding_polygon[index])

        result.append(intersection2[1])

        for point in reversed(poly_line):
            result.append(point)

        result.append(intersection1[1])

        return result




    @staticmethod
    def lineToRay(p0, p1):
        """
        Converts a line segment to a unit vector starting at p0 pointing towards p1.
        @type p0: tuple of (float, float)
        @type p1: tuple of (float, float)
        @rtype: tuple of (float, float)
        """

        x = p1[0] - p0[0]
        y = p1[1] - p0[1]

        length = sqrt(x * x + y * y)

        return x / length, y / length


    @staticmethod
    def rayLineIntersection(point, ray, p1, p2 , flip_ray = False):
        """
        Finds the intersection between the ray starting at point and the line [p1, p2].
        @type point: tuple of (float, float)
        @type ray: tuple of (float, float)
        @type p1: tuple of (float, float)
        @type p2: tuple of (float, float)
        @rtype: tuple of (float, float) or None

        stackoverflow: 563198
        """
        s = (p2[0] - p1[0] , p2[1] - p1[1])
        q = p1
        r = ray
        p = point

        p_m_q = (p[0] - q[0] , p[1] - q[1])
        q_m_p = (q[0] - p[0] , q[1] - p[1])
        r_x_s = r[0] * s[1] - r[1]*s[0]

        q_m_p_x_r = q_m_p[0] * r[1] - q_m_p[1] * r[0]
        q_m_p_x_s = q_m_p[0] * s[1] - q_m_p[1] * s[0]
        
        if abs(r_x_s) < GeometryTools.EPSILON and abs(q_m_p_x_r) < GeometryTools.EPSILON:
            q_m_p_dot_r = q_m_p[0] * r[0] + q_m_p[1] * r[1]
            r_dot_r = r[0] * r[0] + r[1] * r[1]

            p_m_q_dot_s = p_m_q[0] * s[0] + p_m_q[1] * s[1]
            s_dot_s = s[0] * s[0] + s[1] * s[1]

            # Coincident
            if 0 <= q_m_p_dot_r <= r_dot_r:
                return ((p1[0] + p2[0]) / 2 , (p1[1] + p2[1]) / 2)

            # Coincident
            if 0 <= p_m_q_dot_s <= s_dot_s:
                return ((p1[0] + p2[0]) / 2 , (p1[1] + p2[1]) / 2)

            return None
            
            
        if abs(r_x_s) < GeometryTools.EPSILON:
            # Parallell
            return None


        t = 1.0 * q_m_p_x_s / r_x_s
        u = 1.0 * q_m_p_x_r / r_x_s

        if t >= 0 and 0 <= u <= 1:
            x = p[0] + t*r[0]
            y = p[1] + t*r[1]
            
            return x,y

        if flip_ray:
            return GeometryTools.rayLineIntersection( point , (-ray[0] , -ray[1]) , p1 , p2 , False)
        else:
            return None




    @staticmethod
    def rayPolygonIntersections(point, ray, polygon):
        """
        Finds all intersections along the ray with the polygon.
        The returned value is a tuple containing the line segment in the polygon and the intersection coordinate.

        @type point: tuple of (float, float)
        @type ray: tuple of (float, float)
        @type polygon: Polyline or [tuple of (float, float)]
        @rtype: list of tuple of (int, tuple of (float, float))
        """
        results = []
        for index in range(len(polygon) - 1):
            lp1 = polygon[index]
            lp2 = polygon[index + 1]

            intersection = GeometryTools.rayLineIntersection(point, ray, lp1, lp2)
            if intersection is not None:
                results.append((index, intersection))

        return results


    @staticmethod
    def distance(p1,p2):
        if len(p1) != len(p2):
            raise ValueError("Different lenght of objects")
        
        sqr_distance = 0
        for x1,x2 in zip(p1,p2):
            sqr_distance += (x1 - x2) * (x1 - x2)
            
        return sqrt( sqr_distance )
    

    @staticmethod
    def joinPolylines(polyline1 , polyline2):
        """The shortest straight line connecting polyline1 and polyline2.

        The joinPolylines function does not extend the polylines with
        a ray from the end, only the length of the straight line
        connecting the various endpoints is considered. If the two
        polylines already intersect the function returns None.
        """


        if len(polyline1) < 1:
            raise ValueError("Length of polyline must be >= 1")

        if len(polyline2) < 1:
            raise ValueError("Length of polyline must be >= 1")

        if GeometryTools.polylinesIntersect( polyline1 , polyline2):
            return None
            
        p0 = polyline1[0]
        p1 = polyline1[-1]
        pa = polyline2[0]
        pb = polyline2[-1]

        d_list = [ (GeometryTools.distance( p0 , pa ), [p0 , pa]),
                   (GeometryTools.distance( p0 , pb ), [p0 , pb]),
                   (GeometryTools.distance( p1 , pa ), [p1 , pa]),
                   (GeometryTools.distance( p1 , pb ), [p1 , pb]) ]

        d_list.sort( key = lambda x: x[0])
        return d_list[0][1]


    @staticmethod
    def connectPolylines( polyline , target_polyline):
        if GeometryTools.polylinesIntersect( polyline , target_polyline ):
            return None

        if len(polyline) < 2:
            raise ValueError("Polyline must have at least two points")

        d_list = []
        
        p0 = polyline[-1]
        p1 = polyline[-2]
        ray = GeometryTools.lineToRay( p1 , p0 )
        for (index , p) in GeometryTools.rayPolygonIntersections( p0 , ray , target_polyline):
            d_list.append( (GeometryTools.distance( p0 , p) , [p0 , p]) )
        
        p0 = polyline[0]
        p1 = polyline[1]
        ray = GeometryTools.lineToRay( p1 , p0 )
        for (index , p) in GeometryTools.rayPolygonIntersections( p0 , ray , target_polyline):
            d_list.append( (GeometryTools.distance( p0 , p) , [p0 , p]) )

        if len(d_list) == 0:
            raise ValueError("Polyline %s can not be extended to %s" % (polyline.getName() , target_polyline.getName()))

        d_list.sort( key = lambda x: x[0])
        return d_list[0][1]



    @staticmethod
    def nearestPointOnPolyline( p , polyline ):
        if len(polyline) > 1:
            d_list = [ GeometryTools.distance( p  , pi ) for pi in polyline ]
            index0 = d_list.index( min(d_list) )
            p0 = polyline[index0]
            dist0 = d_list[index0]

            dist1 = sys.float_info.max
            dist2 = sys.float_info.max
            intercept1 = None
            intercept2 = None

            index1 = None
            index2 = None
            if index0 > 0:
                index1 = index0 - 1

            if index0 < len(polyline) - 1:
                index2 = index0 + 1

            if not index1 is None:
                p1 = polyline[index1] 
                dy1 = p1[1] - p0[1]
                dx1 = p1[0] - p0[0]
                intercept1 = GeometryTools.rayLineIntersection( p , (dy1 , -dx1) , p0 , p1 , True)
                if intercept1:
                    dist1 = GeometryTools.distance( intercept1 , p )


            if not index2 is None:
                p2 = polyline[index2]
                dy2 = p2[1] - p0[1]
                dx2 = p2[0] - p0[0]
                intercept2 = GeometryTools.rayLineIntersection( p , (dy2 , -dx2) , p0 , p2 , True)
                if intercept2:
                    dist2 = GeometryTools.distance( intercept2 , p )                


            point_list = [ p0 , intercept1 , intercept2 ]
            d_list = [ dist0 , dist1 , dist2 ]
            index = d_list.index( min(d_list) )

            
            return point_list[index]
        else:
            raise ValueError("Polyline must have len() >= 2")




from ecl.util.util import Version
from ecl.util.util import RngAlgTypeEnum, RngInitModeEnum
from ecl.util.util import CTime
from ecl.util.util import PermutationVector
from ecl.util.util import VectorTemplate
from ecl.util.util import DoubleVector
from ecl.util.util import IntVector
from ecl.util.util import BoolVector
from ecl.util.util import TimeVector
from ecl.util.util import StringList
from ecl.util.util import RandomNumberGenerator
from ecl.util.util import LookupTable
from ecl.util.util import Hash, StringHash, DoubleHash, IntegerHash
from ecl.util.util import ThreadPool
from ecl.util.util import installAbortSignals, updateAbortSignals

try:
    from res.util import SubstitutionList
except ImportError:
    pass



#  Copyright (C) 2017 Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import importlib
import os
import sys
import traceback
import unittest
import inspect
import imp

class ImportTestCase(unittest.TestCase):

    def import_file(self, path):
        return imp.load_source( "module", path)

    def import_module(self , module):
        mod = importlib.import_module( module )
        return mod

    def import_package(self, package):
        if "__" in package:
            return True
        module = self.import_module( package )

        path = os.path.dirname( inspect.getfile( module ) )

        for entry in sorted(os.listdir(path)):
            entry_path = os.path.join(path, entry)
            if os.path.isdir( entry_path ):
                module = os.path.basename( entry )
                sub_module = "%s.%s" % (package , module)
                self.import_package( sub_module )
            else:
                module, ext = os.path.splitext( entry )
                if module == "__init__":
                    continue

                if ext == "py":
                    self.import_module("%s.%s" % (package , module))

        return True



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file '__init__.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
ert - Ensemble Reservoir Tool - a package for reservoir modeling.

The ert package itself has no code, but contains several subpackages:

ecl.ecl: Package for working with ECLIPSE files. The far most mature
   package in ert.

ecl.util:

The ert package is based on wrapping the libriaries from the ERT C
code with ctypes; an essential part of ctypes approach is to load the
shared libraries with the ctypes.CDLL() function. The ctypes.CDLL()
function uses the standard methods of the operating system,
i.e. standard locations configured with ld.so.conf and the environment
variable LD_LIBRARY_PATH.

To avoid conflict with other application using the ert libraries the
Python code should be able to locate the shared libraries without
(necessarily) using the LD_LIBRARY_PATH variable. The default
behaviour is to try to load from the library ../../lib64, but by using
the enviornment variable ERT_LIBRARY_PATH you can alter how ert looks
for shared libraries.

   1. By default the code will try to load the shared libraries from
      '../../lib64' relative to the location of this file.

   2. Depending on the value of ERT_LIBRARY_PATH two different
      behaviours can be imposed:

         Existing path: the package will look in the path pointed to
            by ERT_LIBRARY_PATH for shared libraries.

         Arbitrary value: the package will use standard load order for
         the operating system.

If the fixed path, given by the default ../../lib64 or ERT_LIBRARY_PATH
alternative fails, the loader will try the default load behaviour
before giving up completely.
"""
import os.path
import sys

import warnings
warnings.simplefilter('always', DeprecationWarning) # see #1437

from cwrap import load as cwrapload
from cwrap import Prototype

try:
    import ert_site_init
except ImportError:
    pass


required_version_hex = 0x02070000

ecl_lib_path = None
ert_so_version = ""
__version__ = "0.0.0"


# 1. Try to load the __ecl_lib_info module; this module has been
#    configured by cmake during the build configuration process. The
#    module should contain the variable lib_path pointing to the
#    directory with shared object files.
try:
    from .__ecl_lib_info import EclLibInfo
    ecl_lib_path = EclLibInfo.lib_path
    ert_so_version = EclLibInfo.so_version
    __version__ = EclLibInfo.__version__
except ImportError:
    pass
except AttributeError:
    pass


# 2. Using the environment variable ERT_LIBRARY_PATH it is possible to
#    override the default algorithms. If the ERT_LIBRARY_PATH is set
#    to a non existing directory a warning will go to stderr and the
#    setting will be ignored.
env_lib_path = os.getenv("ERT_LIBRARY_PATH")
if env_lib_path:
    if os.path.isdir( env_lib_path ):
        ert_lib_path = os.getenv("ERT_LIBRARY_PATH")
    else:
        sys.stderr.write("Warning: Environment variable ERT_LIBRARY_PATH points to nonexisting directory:%s - ignored" % env_lib_path)


# Check that the final ert_lib_path setting corresponds to an existing
# directory.
if ecl_lib_path:
    if not os.path.isabs(ecl_lib_path):
        ecl_lib_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ecl_lib_path))

    if not os.path.isdir( ecl_lib_path ):
        ecl_lib_path = None

if sys.hexversion < required_version_hex:
    raise Exception("ERT Python requires Python 2.7.")

def load(name):
    try:
        return cwrapload(name, path=ecl_lib_path, so_version=ert_so_version)
    except ImportError:
        # For pip installs, setup.py puts the shared lib in this directory
        own_dir=os.path.dirname(os.path.abspath(__file__))
        sys.stderr.write("Warning: Loading %s from path=%s failed. Trying to load from path=%s...\n" % (name, ecl_lib_path, own_dir))
        return cwrapload(name, path=own_dir, so_version=ert_so_version)

class EclPrototype(Prototype):
    lib = load("libecl")

    def __init__(self, prototype, bind=True):
        super(EclPrototype, self).__init__(EclPrototype.lib, prototype, bind=bind)

from .ecl_type import EclTypeEnum, EclDataType
from .ecl_util import EclFileEnum, EclFileFlagEnum, EclPhaseEnum, EclUnitTypeEnum , EclUtil

from .util.util import EclVersion
from .util.util import updateAbortSignals

updateAbortSignals( )

def root():
    """
    Will print the filesystem root of the current ert package.
    """
    return os.path.abspath( os.path.join( os.path.dirname( __file__ ) , "../"))




#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'int_vector.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl import EclPrototype
from ecl.util.util import VectorTemplate


class IntVector(VectorTemplate):
    default_format       = "%d"

    _alloc               = EclPrototype("void*  int_vector_alloc( int , int )" , bind = False)
    _create_active_list  = EclPrototype("int_vector_obj string_util_alloc_active_list( char*)" , bind = False)
    _create_value_list   = EclPrototype("int_vector_obj string_util_alloc_value_list( char*)" , bind = False)
    _alloc_copy          = EclPrototype("int_vector_obj int_vector_alloc_copy( int_vector )")
    _strided_copy        = EclPrototype("int_vector_obj int_vector_alloc_strided_copy( int_vector , int , int , int)")
    _free                = EclPrototype("void   int_vector_free( int_vector )")
    _iget                = EclPrototype("int    int_vector_iget( int_vector , int )")
    _safe_iget           = EclPrototype("int    int_vector_safe_iget( int_vector , int )")
    _iset                = EclPrototype("int    int_vector_iset( int_vector , int , int)")
    _size                = EclPrototype("int    int_vector_size( int_vector )")
    _append              = EclPrototype("void   int_vector_append( int_vector , int )")
    _idel_block          = EclPrototype("void   int_vector_idel_block( int_vector , int , int )")
    _pop                 = EclPrototype("int    int_vector_pop( int_vector )")
    _idel                = EclPrototype("void   int_vector_idel( int_vector , int )")
    _insert              = EclPrototype("void   int_vector_insert( int_vector , int , int)")
    _lshift              = EclPrototype("void   int_vector_lshift( int_vector , int )")
    _rshift              = EclPrototype("void   int_vector_rshift( int_vector , int )")
    _fprintf             = EclPrototype("void   int_vector_fprintf( int_vector , FILE , char* , char*)")
    _sort                = EclPrototype("void   int_vector_sort( int_vector )")
    _rsort               = EclPrototype("void   int_vector_rsort( int_vector )")
    _reset               = EclPrototype("void   int_vector_reset( int_vector )")
    _set_read_only       = EclPrototype("void   int_vector_set_read_only( int_vector , bool )")
    _get_read_only       = EclPrototype("bool   int_vector_get_read_only( int_vector )")
    _get_max             = EclPrototype("int    int_vector_get_max( int_vector )")
    _get_min             = EclPrototype("int    int_vector_get_min( int_vector )")
    _get_max_index       = EclPrototype("int    int_vector_get_max_index( int_vector , bool)")
    _get_min_index       = EclPrototype("int    int_vector_get_min_index( int_vector , bool)")
    _shift               = EclPrototype("void   int_vector_shift( int_vector , int )")
    _scale               = EclPrototype("void   int_vector_scale( int_vector , int )")
    _div                 = EclPrototype("void   int_vector_div( int_vector , int )")
    _inplace_add         = EclPrototype("void   int_vector_inplace_add( int_vector , int_vector )")
    _inplace_mul         = EclPrototype("void   int_vector_inplace_mul( int_vector , int_vector )")
    _assign              = EclPrototype("void   int_vector_set_all( int_vector , int)")
    _memcpy              = EclPrototype("void   int_vector_memcpy(int_vector , int_vector )")
    _set_default         = EclPrototype("void   int_vector_set_default( int_vector , int)")
    _get_default         = EclPrototype("int    int_vector_get_default( int_vector )")
    _element_size        = EclPrototype("int    int_vector_element_size( int_vector )")

    _permute             = EclPrototype("void int_vector_permute(int_vector, permutation_vector)")
    _sort_perm           = EclPrototype("permutation_vector_obj int_vector_alloc_sort_perm(int_vector)")
    _rsort_perm          = EclPrototype("permutation_vector_obj int_vector_alloc_rsort_perm(int_vector)")
    _contains            = EclPrototype("bool int_vector_contains(int_vector, int)")
    _select_unique       = EclPrototype("void int_vector_select_unique(int_vector)")
    _element_sum         = EclPrototype("int int_vector_sum(int_vector)")
    _get_data_ptr        = EclPrototype("int* int_vector_get_ptr(int_vector)")
    _count_equal         = EclPrototype("int int_vector_count_equal(int_vector, int)")
    _init_range          = EclPrototype("void int_vector_init_range(int_vector, int , int , int)")
    _init_linear         = EclPrototype("bool int_vector_init_linear(int_vector, int, int, int)")
    _equal               = EclPrototype("bool int_vector_equal(int_vector, int_vector)")
    _first_eq            = EclPrototype("int int_vector_first_equal(int_vector, int_vector, int)")
    _first_neq           = EclPrototype("int int_vector_first_not_equal(int_vector, int_vector, int)")

    def __init__(self, default_value=0, initial_size=0):
        super(IntVector, self).__init__(default_value, initial_size)

    @classmethod
    def active_list(cls, range_string):
        """Will create a IntVector instance with the values from @range_string.

        The range_string input should be of the type "1,3-5,9,17",
        i.e. integer values separated by commas, and dashes to
        represent ranges. If the input string contains ANY invalid
        characters the returned active list will be empty:

           "1,4-7,10"  =>  {1,4,5,6,7,10}
           "1,4-7,10X" =>  {}

        The empty list will evaluate to false. The values in the input
        string are meant to indicate "active values", i.e. the output
        values are sorted and repeated values are only counted once:

           "1,1,7,2" => {1,2,7}

        """
        return cls._create_active_list(range_string)

    @classmethod
    def valueList(cls , range_string):
        """Will create a IntVecter of all the values in the @range_string.

        Will not sort the values, and not uniquiefy - in contrast to
        the active_list() method.

        """
        return cls._create_value_list(range_string)


    def count(self, value):
        """ @rtype: int """
        return self._count_equal(value)



#!/usr/bin/env python
#  Copyright (C) 2017 Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

import sys
import fnmatch
import os
import unittest
try:
    from pylint import epylint as lint
except ImportError:
    sys.stderr.write("Could not import pylint module - lint based testing will be skipped\n")
    lint = None


class LintTestCase(unittest.TestCase):
    """This class is a test case for linting."""

    LINT_ARGS = ['-d', 'R,C,W'] + \
                ['--extension-pkg-whitelist=numpy']


    @staticmethod
    def _get_lintable_files(paths, whitelist=()):
        """Recursively traverses all folders in paths for *.py files"""
        matches = []
        for folder in paths:
            for root, _, filenames in os.walk(folder):
                for filename in fnmatch.filter(filenames, '*.py'):
                    if filename not in whitelist:
                        matches.append(os.path.join(root, filename))
        return matches


    def assertLinted(self, paths, whitelist=()):  # noqa
        """Takes a path to a folder or a list of paths to folders and recursively finds
        all *.py files within that folder except the ones with filenames in whitelist.

        Will assert lint.lint(fname) == 0 for every *.py file found.
        """
        if lint is None:
            self.skipTest("pylint not installed")

        if isinstance(paths, str):
            paths = [paths]
        files = self._get_lintable_files(paths, whitelist=whitelist)
        for f in files:
            self.assertEqual(0, lint.lint(f, self.LINT_ARGS), 'Linting required for %s' % f)



#  Copyright (C) 2016  Statoil ASA, Norway.
#
#  This file is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from cwrap import BaseCClass
from ecl.util.util import monkey_the_camel
from ecl import EclPrototype


class EclSMSPECNode(BaseCClass):
    """
    Small class with some meta information about a summary variable.

    The summary variables have different attributes, like if they
    represent a total quantity, a rate or a historical quantity. These
    quantities, in addition to the underlying values like WGNAMES,
    KEYWORD and NUMS taken from the the SMSPEC file are stored in this
    structure.
    """
    TYPE_NAME = "smspec_node"
    _node_is_total      = EclPrototype("bool smspec_node_is_total( smspec_node )")
    _node_is_historical = EclPrototype("bool smspec_node_is_historical( smspec_node )")
    _node_is_rate       = EclPrototype("bool smspec_node_is_rate( smspec_node )")
    _node_unit          = EclPrototype("char* smspec_node_get_unit( smspec_node )")
    _node_wgname        = EclPrototype("char* smspec_node_get_wgname( smspec_node )")
    _node_keyword       = EclPrototype("char* smspec_node_get_keyword( smspec_node )")
    _node_num           = EclPrototype("int   smspec_node_get_num( smspec_node )")
    _node_need_num      = EclPrototype("bool  smspec_node_need_nums( smspec_node )")
    _gen_key1           = EclPrototype("char* smspec_node_get_gen_key1( smspec_node )")
    _gen_key2           = EclPrototype("char* smspec_node_get_gen_key2( smspec_node )")
    _var_type           = EclPrototype("ecl_sum_var_type smspec_node_get_var_type( smspec_node )")
    _cmp                = EclPrototype("int smspec_node_cmp( smspec_node , smspec_node)")
    _get_default        = EclPrototype("float smspec_node_get_default(smspec_node)")

    def __init__(self):
        super(EclSMSPECNode, self).__init__(0) # null pointer
        raise NotImplementedError("Class can not be instantiated directly!")

    def cmp(self, other):
        if isinstance(other, EclSMSPECNode):
            return self._cmp( other )
        else:
            raise TypeError("Other argument must be of type EclSMSPECNode")


    def __lt__(self , other):
        return self.cmp( other ) < 0


    def __gt__(self , other):
        return self.cmp( other ) > 0


    def __eq__(self , other):
        return self.cmp( other ) == 0


    def __hash__(self , other):
        return hash(self._gen_key1( ))


    @property
    def unit(self):
        """
        Returns the unit of this node as a string.
        """
        return self._node_unit( )

    @property
    def wgname(self):
        """
        Returns the WGNAME property for this node.

        Many variables do not have the WGNAME property, i.e. the field
        related variables like FOPT and the block properties like
        BPR:10,10,10. For these variables the function will return
        None, and not the ECLIPSE dummy value: ":+:+:+:+".
        """
        return self._node_wgname( )


    @property
    def keyword(self):
        """
        Returns the KEYWORD property for this node.

        The KEYWORD property is the main classification property in
        the ECLIPSE SMSPEC file. The properties of a variable can be
        read from the KEWYORD value; see table 3.4 in the ECLIPSE file
        format reference manual.
        """
        return self._node_keyword( )

    @property
    def num(self):
        return self.getNum( )


    @property
    def default(self):
        """Will return the default value for this key.

        The default value is typically used when fetching values from a
        historical case, when the key is only present in the restarted case.
        The default value is also used to initialize the PARAMS vector when
        writing to file.
        """
        return self._get_default()


    def get_key1(self):
        """
        Returns the primary composite key, i.e. like 'WOPR:OPX' for this
        node.
        """
        return self._gen_key1( )


    def get_key2(self):
        """Returns the secondary composite key for this node.

        Most variables have only one composite key, but in particular
        nodes which involve (i,j,k) coordinates will contain two
        forms:

            getKey1()  =>  "BPR:10,11,6"
            getKey2()  =>  "BPR:52423"

        Where the '52423' in getKey2() corresponds to i + j*nx +
        k*nx*ny.
        """
        return self._gen_key2( )


    def var_type(self):
        return self._var_type( )


    def get_num(self):
        """
        Returns the NUMS value for this keyword; or None.

        Many of the summary keywords have an integer stored in the
        vector NUMS as an attribute, i.e. the block properties have
        the global index of the cell in the nums vector. If the
        variable in question makes use of the NUMS value this property
        will return the value, otherwise it will return None:

           sum.smspec_node("FOPT").num     => None
           sum.smspec_node("BPR:1000").num => 1000

        """
        if self._node_need_num( ):
            return self._node_num( )
        else:
            return None

    def is_rate(self):
        """
        Will check if the variable in question is a rate variable.

        The conecpt of rate variabel is important (internally) when
        interpolation values to arbitrary times.
        """
        return self._node_is_rate()


    def is_total(self):
        """
        Will check if the node corresponds to a total quantity.

        The question of whether a variable corresponds to a 'total'
        quantity or not can be interesting for e.g. interpolation
        purposes. The actual question whether a quantity is total or
        not is based on a hardcoded list in smspec_node_set_flags() in
        smspec_node.c; this list again is based on the tables 2.7 -
        2.11 in the ECLIPSE fileformat documentation.
        """
        return self._node_is_total( )


    def is_historical(self):
        """
        Checks if the key corresponds to a historical variable.

        The check is only based on the last character; all variables
        ending with 'H' are considered historical.
        """
        return self._node_is_historical( )


monkey_the_camel(EclSMSPECNode, 'getKey1', EclSMSPECNode.get_key1)
monkey_the_camel(EclSMSPECNode, 'getKey2', EclSMSPECNode.get_key2)
monkey_the_camel(EclSMSPECNode, 'varType', EclSMSPECNode.var_type)
monkey_the_camel(EclSMSPECNode, 'getNum', EclSMSPECNode.get_num)
monkey_the_camel(EclSMSPECNode, 'isRate', EclSMSPECNode.is_rate)
monkey_the_camel(EclSMSPECNode, 'isTotal', EclSMSPECNode.is_total)
monkey_the_camel(EclSMSPECNode, 'isHistorical', EclSMSPECNode.is_historical)



#  Copyright (C) 2015  Statoil ASA, Norway.
#
#  The file 'ecl_cmp.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl.util.util import monkey_the_camel
from ecl.summary import EclSum

class EclCase(object):

    def __init__(self, case):
        self.case = case

        self.grid = None
        self.restart = None
        self.init = None
        self.summary = None

        self.loadSummary()


    def __contains__(self, key):
        return key in self.summary


    def keys(self):
        return self.summary.keys()


    def wells(self):
        return self.summary.wells()

    def load_summary(self):
        self.summary = EclSum(self.case)


    def start_time_equal(self, other):
        if self.summary.getDataStartTime() == other.summary.getDataStartTime():
            return True
        else:
            return False

    def end_time_equal(self, other):
        if self.summary.getEndTime() == other.summary.getEndTime():
            return True
        else:
            return False


    def cmp_summary_vector(self, other, key, sample=100):
        if key in self and key in other:
            days_total = min(self.summary.getSimulationLength(), other.summary.getSimulationLength())
            dt = days_total / (sample - 1)
            days = [ x * dt for x in range(sample) ]

            ref_data = self.summary.get_interp_vector(key, days_list=days)
            test_data = other.summary.get_interp_vector(key, days_list=days)
            diff_data = ref_data - test_data

            ref_sum = sum(ref_data)
            diff_sum = sum(abs(diff_data))
            return (diff_sum, ref_sum)
        else:
            raise KeyError("Key:%s was not present in both cases" % key)





class EclCmp(object):

    def __init__(self, test_case, ref_case):
        """Class to compare to simulation cases with Eclipse formatted result files.

        The first argument is supposed to be the test_case and the
        second argument is the reference case. The arguemnts should be
        the basenames of the simulation, with an optional path
        prefix - an extension is accepted, but will be ignored.

        The constructor will start be calling the method initCheck()
        to check that the two cases are 'in the same ballpark'.
        """
        self.test_case = EclCase(test_case)
        self.ref_case = EclCase(ref_case)

        self.initCheck()


    def init_check(self):
        """A crude initial check to verify that the cases can be meaningfully
        compared.
        """
        if not self.test_case.startTimeEqual(self.ref_case):
            raise ValueError("The two cases do not start at the same time - can not be compared")


    def has_summary_vector(self, key):
        """
        Will check if both test and refernce have @key.
        """
        return (key in self.test_case, key in self.ref_case)


    def end_time_equal(self):
        """
        Will check that ref_case and test_case are equally long.
        """
        return self.test_case.endTimeEqual(self.ref_case)



    def cmp_summary_vector(self, key, sample=100):
        """Will compare the summary vectors according to @key.

        The comparison is based on evaluating the integrals:

           I0 = \int R(t) dt

           delta = \int | R(t) - T(t)| dt

        numericall. R(t) is the reference solution and T(t) is
        testcase solution. The return value is a tuple:

             (delta, I0)

        So that a natural way to evaluate the check for numerical
        equality, based on the relative error could be:

           delta, scale = ecl_cmp.cmpSummaryVector("WWCT:OP_1")

           if delta/scale < 0.0001:
               print("Equal enough")
           else:
               print("Different ..")

        The upper limit for the integrals is:

           max(length(ref_case), length(test_case))

        meaning that two simulations which don't have the same
        end-time will compare as equal if they compare equal in the
        common part. If that is not OK you should call the
        endTimeEqual() method independently.
        """
        return self.test_case.cmpSummaryVector(self.ref_case, key, sample=sample)


    def test_keys(self):
        """
        Will return a list of summary keys in the test case.
        """
        return self.test_case.keys()


    def test_wells(self):
        """
        Will return a list of wells keys in the test case.
        """
        return self.test_case.wells()



monkey_the_camel(EclCase, 'loadSummary', EclCase.load_summary)
monkey_the_camel(EclCase, 'startTimeEqual', EclCase.start_time_equal)
monkey_the_camel(EclCase, 'endTimeEqual', EclCase.end_time_equal)
monkey_the_camel(EclCase, 'cmpSummaryVector', EclCase.cmp_summary_vector)

monkey_the_camel(EclCmp, 'initCheck', EclCmp.init_check)
monkey_the_camel(EclCmp, 'hasSummaryVector', EclCmp.has_summary_vector)
monkey_the_camel(EclCmp, 'endTimeEqual', EclCmp.end_time_equal)
monkey_the_camel(EclCmp, 'cmpSummaryVector', EclCmp.cmp_summary_vector)
monkey_the_camel(EclCmp, 'testKeys', EclCmp.test_keys)
monkey_the_camel(EclCmp, 'testWells', EclCmp.test_wells)



import os

try:
    from unittest2 import TestLoader, TextTestRunner
except ImportError:
    from unittest import TestLoader, TextTestRunner


class ErtTestRunner(object):

    @staticmethod
    def runTestSuite(tests , test_verbosity = 3):
        test_runner = TextTestRunner(verbosity=test_verbosity)
        result = test_runner.run(tests)

        return result.wasSuccessful()


    @staticmethod
    def findTestsInDirectory(path, recursive=True , pattern = "test*.py"):
        loader = TestLoader()
        test_suite = loader.discover(path , pattern = pattern)
        
        for (root, dirnames, filenames) in os.walk( path ):
            for directory in dirnames:
                test_suite.addTests(ErtTestRunner.findTestsInDirectory(os.path.join(root, directory), recursive , pattern))

        return test_suite


    @staticmethod
    def runTestsInDirectory(path=".", recursive=True, test_verbosity=3):
        test_suite = ErtTestRunner.findTestsInDirectory(path, recursive)
        return ErtTestRunner.runTestSuite(test_suite)
        

    @staticmethod
    def runTestsInClass(classpath, test_verbosity=3):
        klass = ErtTestRunner.importClass(classpath)
        loader = TestLoader()
        tests = loader.loadTestsFromTestCase(klass)
        testRunner = TextTestRunner(verbosity=test_verbosity)
        testRunner.run(tests)


    @staticmethod
    def importClass(classpath):
        dot = classpath.rfind(".")
        class_name = classpath[dot + 1:]
        try:
            m = __import__(classpath[0:dot], globals(), locals(), [class_name])
            return getattr(m, class_name)
        except ImportError:
            print("Failed to import: %s" % classpath)
            raise


    @staticmethod
    def getTestsFromTestClass(test_class_path, argv=None):
        klass = ErtTestRunner.importClass(test_class_path)
        klass.argv = argv
        loader = TestLoader()
        return loader.loadTestsFromTestCase(klass)



#  Copyright (C) 2011  Statoil ASA, Norway.
#
#  The file 'ecl_grav.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.
"""
Calculate dynamic change in gravitational strength.

The ecl_grav module contains functionality to load time-lapse ECLIPSE
results and calculate the change in gravitational strength between the
different surveys. The implementation is a thin wrapper around the
ecl_grav.c implementation in the libecl library.
"""
from cwrap import BaseCClass

from ecl import EclPrototype
from ecl.util.util import monkey_the_camel
from ecl import EclPhaseEnum
import ecl.eclfile

class EclGrav(BaseCClass):
    """
    Holding ECLIPSE results for calculating gravity changes.

    The EclGrav class is a collection class holding the results from
    ECLIPSE forward modelling of gravity surveys. Observe that the
    class is focused on the ECLIPSE side of things, and does not have
    any notion of observed values or measurement locations; that
    should be handled by the scope using the EclGrav class.

    Typical use of the EclGrav class involves the following steps:

      1. Create the EclGrav instance.
      2. Add surveys with the add_survey_XXXX() methods.
      3. Evalute the gravitational response with the eval() method.
    """
    TYPE_NAME = "ecl_grav"
    _grav_alloc = EclPrototype("void* ecl_grav_alloc(ecl_grid, ecl_file)", bind=False)
    _free = EclPrototype("void ecl_grav_free(ecl_grav)")
    _add_survey_RPORV = EclPrototype("void*  ecl_grav_add_survey_RPORV(ecl_grav, char*, ecl_file_view)")
    _add_survey_PORMOD = EclPrototype("void*  ecl_grav_add_survey_PORMOD(ecl_grav, char*, ecl_file_view)")
    _add_survey_FIP = EclPrototype("void*  ecl_grav_add_survey_FIP(ecl_grav, char*, ecl_file_view)")
    _add_survey_RFIP = EclPrototype("void*  ecl_grav_add_survey_RFIP(ecl_grav, char*, ecl_file_view)")
    _new_std_density = EclPrototype("void ecl_grav_new_std_density(ecl_grav, int, double)")
    _add_std_density = EclPrototype("void ecl_grav_add_std_density(ecl_grav, int, int, double)")
    _eval = EclPrototype("double ecl_grav_eval(ecl_grav, char*, char*, ecl_region, double, double, double, int)")



    def __init__(self, grid, init_file):
        """
        Creates a new EclGrav instance.

        The input arguments @grid and @init_file should be instances
        of EclGrid and EclFile respectively.
        """
        self.init_file = init_file   # Inhibit premature garbage collection of init_file

        c_ptr = self._grav_alloc(grid, init_file)
        super(EclGrav, self).__init__(c_ptr)

        self.dispatch = {"FIP"    : self.add_survey_FIP,
                         "RFIP"   : self.add_survey_RFIP,
                         "PORMOD" : self.add_survey_PORMOD,
                         "RPORV"  : self.add_survey_RPORV}


    def add_survey_RPORV(self, survey_name, restart_view):
        """
        Add new survey based on RPORV keyword.

        Add a new survey; in this context a survey is the state of
        reservoir, i.e. an ECLIPSE restart file. The @survey_name
        input argument will be used when refering to this survey at a
        later stage. The @restart_view input argument should be an
        EclFile instance with data from one report step. A typical way
        to load the @restart_view argument is:

           import datetime
           from ecl.ecl import EclRestartFile
           ...
           ...
           date = datetime.datetime(year, month, day)
           rst_file = EclRestartFile("ECLIPSE.UNRST")
           restart_view1 = rst_file.restartView(sim_time=date)
           restart_view2 = rst_file.restartView(report_step=67)

        The pore volume of each cell will be calculated based on the
        RPORV keyword from the restart files. The methods
        add_survey_PORMOD() and add_survey_FIP() are alternatives
        which are based on other keywords.
        """
        self._add_survey_RPORV(survey_name, restart_view)

    def add_survey_PORMOD(self, survey_name, restart_view):
        """
        Add new survey based on PORMOD keyword.

        The pore volum is calculated from the initial pore volume and
        the PORV_MOD keyword from the restart file; see
        add_survey_RPORV() for further details.
        """
        self._add_survey_PORMOD(survey_name, restart_view)

    def add_survey_FIP(self, survey_name, restart_view):
        """
        Add new survey based on FIP keywords.

        This method adds a survey as add_survey_RPORV() and
        add_survey_PORMOD; but the mass content in each cell is
        calculated based on the FIPxxx keyword along with the mass
        density at standard conditions of the respective phases.

        The mass density at standard conditions must be specified with
        the new_std_density() (and possibly also add_std_density())
        method before calling the add_survey_FIP() method.
        """
        self._add_survey_FIP(survey_name, restart_view)

    def add_survey_RFIP(self, survey_name, restart_view):
        """
        Add new survey based on RFIP keywords.

        This method adds a survey as add_survey_RPORV() and
        add_survey_PORMOD; but the mass content in each cell is
        calculated based on the RFIPxxx keyword along with the
        per-cell mass density of the respective phases.
        """
        self._add_survey_RFIP(survey_name, restart_view)

    def add_survey(self, name, restart_view, method):
        method = self.dispatch[ method ]
        return method(name, restart_view)

    def eval(self, base_survey, monitor_survey, pos, region=None,
             phase_mask=EclPhaseEnum.ECL_OIL_PHASE + EclPhaseEnum.ECL_GAS_PHASE + EclPhaseEnum.ECL_WATER_PHASE):
        """
        Calculates the gravity change between two surveys.

        This is the method everything is leading up to; will calculate
        the change in gravitational strength, in units of micro Gal,
        between the two surveys named @base_survey and
        @monitor_survey.

        The monitor survey can be 'None' - the resulting answer has
        nothing whatsovever to do with gravitation, but can be
        interesting to determine the numerical size of the quantities
        which are subtracted in a 4D study.

        The @pos argument should be a tuple of three elements with the
        (utm_x, utm_y, depth) position where we want to evaluate the
        change in gravitational strength.

        If supplied the optional argument @region should be an
        EclRegion() instance; this region will be used to limit the
        part of the reserviour included in the gravity calculations.

        The optional argument @phase_mask is an integer flag to
        indicate which phases you are interested in. It should be a
        sum of the relevant integer constants 'ECL_OIL_PHASE',
        'ECL_GAS_PHASE' and 'ECL_WATER_PHASE'.
        """
        return self._eval(base_survey, monitor_survey, region, pos[0], pos[1], pos[2], phase_mask)


    def new_std_density(self, phase_enum, default_density):
        """
        Adds a new phase with a corresponding density.

        @phase_enum is one of the integer constants ECL_OIL_PHASE,
        ECL_GAS_PHASE or ECL_WATER_PHASE, all available in the
        ecl_util and also ecl modules.

        @default_density is the density, at standard conditions, for
        this particular phase. By default @default_density will be
        used for all the cells in the model; by using the
        add_std_density() method you can specify different densities
        for different PVT regions.

        The new_std_density() and add_std_density() methods must be
        used before you use the add_survey_FIP() method to add a
        survey based on the FIP keyword.
        """
        self._new_std_density(phase_enum, default_density)

    def add_std_density(self, phase_enum, pvtnum, density):
        """
        Add standard conditions density for PVT region @pvtnum.

        The new_std_density() method will add a standard conditions
        density which applies to all cells in the model. Using the
        add_std_density() method it is possible to add standard
        conditions densities on a per PVT region basis. You can add
        densities for as many PVT regions as you like, and then fall
        back to the default density for the others.

        The new_std_density() method must be called before calling the
        add_std_density() method.

        The new_std_density() and add_std_density() methods must be
        used before you use the add_survey_FIP() method to add a
        survey based on the FIP keyword.
        """
        self._add_std_density(phase_enum, pvtnum, density)


    def free(self):
        self._free()

monkey_the_camel(EclGrav, 'addSurvey', EclGrav.add_survey)



from cwrap import BaseCClass
from ecl import EclPrototype
from ecl.well import WellState

class WellTimeLine(BaseCClass):
    TYPE_NAME = "well_time_line"
    _size = EclPrototype("int well_ts_get_size(well_time_line)")
    _name = EclPrototype("char* well_ts_get_name(well_time_line)")
    _iget = EclPrototype("well_state_ref well_ts_iget_state(well_time_line, int)")

    def __init__(self):
        raise NotImplementedError("Class can not be instantiated directly")

    def getName(self):
        return self._name()

    def __len__(self):
        """ @rtype: int """
        return self._size()


    def __getitem__(self, index):
        """
         @type index: int
         @rtype: WellState
        """

        if index < 0:
            index += len(self)

        if not 0 <= index < len(self):
            raise IndexError("Index must be in range 0 <= %d < %d" % (index, len(self)))

        return self._iget(index).setParent(self)

    def free(self):
        pass

    def __repr__(self):
        n = self.getName()
        l = len(self)
        cnt = 'name = %s, size = %d' % (n,l)
        return self._create_repr(cnt)



#  Copyright (C) 2014  Statoil ASA, Norway.
#
#  The file 'vector_template.py' is part of ERT - Ensemble based Reservoir Tool.
#
#  ERT is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ERT is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http://www.gnu.org/licenses/gpl.html>
#  for more details.

from ecl import EclPrototype
from ecl.util.util import VectorTemplate


class BoolVector(VectorTemplate):
    default_format       = "%8d"

    _alloc               = EclPrototype("void*  bool_vector_alloc( int , bool )" , bind = False)
    _create_active_mask  = EclPrototype("bool_vector_obj string_util_alloc_active_mask( char* )" , bind = False)
    _active_list         = EclPrototype("int_vector_obj bool_vector_alloc_active_list(bool_vector)", bind = False)
    _alloc_copy          = EclPrototype("bool_vector_obj bool_vector_alloc_copy( bool_vector )")
    _update_active_mask  = EclPrototype("bool string_util_update_active_mask(char*, bool_vector)" , bind = False)

    _strided_copy        = EclPrototype("bool_vector_obj bool_vector_alloc_strided_copy( bool_vector , int , int , int)")
    _free                = EclPrototype("void   bool_vector_free( bool_vector )")
    _iget                = EclPrototype("bool   bool_vector_iget( bool_vector , int )")
    _safe_iget           = EclPrototype("bool   bool_vector_safe_iget( bool_vector , int )")
    _iset                = EclPrototype("void   bool_vector_iset( bool_vector , int , bool)")
    _size                = EclPrototype("int    bool_vector_size( bool_vector )")
    _append              = EclPrototype("void   bool_vector_append( bool_vector , bool )")
    _idel_block          = EclPrototype("void   bool_vector_idel_block( bool_vector , bool , bool )")
    _idel                = EclPrototype("void   bool_vector_idel( bool_vector , int )")
    _pop                 = EclPrototype("bool   bool_vector_pop( bool_vector )")
    _lshift              = EclPrototype("void   bool_vector_lshift( bool_vector , int )")
    _rshift              = EclPrototype("void   bool_vector_rshift( bool_vector , int )")
    _insert              = EclPrototype("void   bool_vector_insert( bool_vector , int , bool)")
    _fprintf             = EclPrototype("void   bool_vector_fprintf( bool_vector , FILE , char* , char*)")
    _sort                = EclPrototype("void   bool_vector_sort( bool_vector )")
    _rsort               = EclPrototype("void   bool_vector_rsort( bool_vector )")
    _reset               = EclPrototype("void   bool_vector_reset( bool_vector )")
    _set_read_only       = EclPrototype("void   bool_vector_set_read_only( bool_vector , bool )")
    _get_read_only       = EclPrototype("bool   bool_vector_get_read_only( bool_vector )")
    _get_max             = EclPrototype("bool   bool_vector_get_max( bool_vector )")
    _get_min             = EclPrototype("bool   bool_vector_get_min( bool_vector )")
    _get_max_index       = EclPrototype("int    bool_vector_get_max_index( bool_vector , bool)")
    _get_min_index       = EclPrototype("int    bool_vector_get_min_index( bool_vector , bool)")
    _shift               = EclPrototype("void   bool_vector_shift( bool_vector , bool )")
    _scale               = EclPrototype("void   bool_vector_scale( bool_vector , bool )")
    _div                 = EclPrototype("void   bool_vector_div( bool_vector , bool )")
    _inplace_add         = EclPrototype("void   bool_vector_inplace_add( bool_vector , bool_vector )")
    _inplace_mul         = EclPrototype("void   bool_vector_inplace_mul( bool_vector , bool_vector )")
    _assign              = EclPrototype("void   bool_vector_set_all( bool_vector , bool)")
    _memcpy              = EclPrototype("void   bool_vector_memcpy(bool_vector , bool_vector )")
    _set_default         = EclPrototype("void   bool_vector_set_default( bool_vector , bool)")
    _get_default         = EclPrototype("bool   bool_vector_get_default( bool_vector )")
    _element_size        = EclPrototype("int    bool_vector_element_size( bool_vector )")

    _permute             = EclPrototype("void bool_vector_permute(bool_vector, permutation_vector)")
    _sort_perm           = EclPrototype("permutation_vector_obj bool_vector_alloc_sort_perm(bool_vector)")
    _rsort_perm          = EclPrototype("permutation_vector_obj bool_vector_alloc_rsort_perm(bool_vector)")

    _contains            = EclPrototype("bool bool_vector_contains(bool_vector, bool)")
    _select_unique       = EclPrototype("void bool_vector_select_unique(bool_vector)")
    _element_sum         = EclPrototype("bool bool_vector_sum(bool_vector)")
    _get_data_ptr        = EclPrototype("bool* bool_vector_get_ptr(bool_vector)")
    _count_equal         = EclPrototype("int bool_vector_count_equal(bool_vector, bool)")
    _init_linear         = None
    _equal               = EclPrototype("bool bool_vector_equal(bool_vector, bool_vector)")
    _first_eq            = EclPrototype("int bool_vector_first_equal(bool_vector, bool_vector, int)")
    _first_neq           = EclPrototype("int bool_vector_first_not_equal(bool_vector, bool_vector, int)")


    def __init__(self, default_value=False, initial_size=0):
        super(BoolVector, self).__init__(default_value, initial_size)

    def count(self, value=True):
        """ @rtype: int """
        return self._count_equal(self, value)

    @classmethod
    def createActiveMask(cls, range_string):
        """
        Will create a BoolVector instance with the values from @range_string.

        The range_string input should be of the type "1,3-5,9,17",
        i.e. integer values separated by commas, and dashes to
        represent ranges. If the input string contains ANY invalid
        characters the returned active list will be empty:

           "1,4-7,10"  =>  {F,T,F,F,T,T,T,T,F,F,T}
           "1,4-7,10X" =>  {}

        The empty list will evaluate to false
        @rtype: BoolVector
        """
        return cls._create_active_mask(range_string)

    def updateActiveMask(self, range_string):
        """
        Updates a bool vector based on a range string.
        @type range_string: str
        @type bool_vector: BoolVector
        @rtype: bool
        """
        return self._update_active_mask(range_string , self)

    @classmethod
    def createFromList(cls, size, source_list):
        """
        Allocates a bool vector from a Python list of indexes
        @rtype: BoolVector
        """
        bool_vector = BoolVector(False, size)

        for index in source_list:
            index = int(index)
            bool_vector[index] = True

        return bool_vector

    def createActiveList(self):
        """ @rtype: ecl.util.IntVector """
        return self._active_list(self)

    def _tostr(self, arr = None):
        if arr is None:
            arr = self
        return "".join(['1' if x else '0' for x in arr])

    def __repr__(self):
        """Will return BoolVector(size = 4, content = "0010") at 0x1729 and
        if size > 10, will return content = "0001...100", i.e., |content|<=10.
        """
        cnt = ''
        ls = len(self)
        if ls <= 20:
            cnt = self._tostr()
        else:
            a,b  = self[:9], self[-8:]
            cnt  = self._tostr(a)
            cnt += "..."
            cnt += self._tostr(b)
        return 'BoolVector(size = %d, content = "%s") %s' % (ls, cnt, self._ad_str())


    @classmethod
    def create_linear(cls, start_value, end_value, num_values):
        raise NotImplementedError("The init_linear method does not make sense for bool vectors")
